# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from unittest import mock

import ddt
from oslo_utils import uuidutils

from tacker import context
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import lcm_op_occ_utils as lcmocc_utils
from tacker.sol_refactored.conductor import conductor_v2
from tacker.sol_refactored.conductor import vnflcm_driver_v2
from tacker.sol_refactored.nfvo import nfvo_client
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields
from tacker.tests.unit.db import base as db_base


@ddt.ddt
class TestConductorV2(db_base.SqlTestCase):

    def setUp(self):
        super(TestConductorV2, self).setUp()
        objects.register_all()
        self.conductor = conductor_v2.ConductorV2()
        self.context = context.get_admin_context()

    def _create_inst_and_lcmocc(self,
            op_state=fields.LcmOperationStateType.STARTING):
        inst = objects.VnfInstanceV2(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfdId=uuidutils.generate_uuid(),
            vnfProvider='provider',
            vnfProductName='product name',
            vnfSoftwareVersion='software version',
            vnfdVersion='vnfd version',
            instantiationState='NOT_INSTANTIATED'
        )

        req = {"flavourId": "simple"}  # instantiate request
        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=op_state,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            vnfInstanceId=inst.id,
            operation=fields.LcmOperationType.INSTANTIATE,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req)

        inst.create(self.context)
        lcmocc.create(self.context)

        return lcmocc

    def _make_grant_req_and_grant(self, lcmocc):
        grant_req = objects.GrantRequestV1(
            # required fields
            vnfInstanceId=lcmocc.vnfInstanceId,
            vnfLcmOpOccId=lcmocc.id,
            vnfdId=uuidutils.generate_uuid(),
            operation=lcmocc.operation,
            isAutomaticInvocation=lcmocc.isAutomaticInvocation
        )
        grant = objects.GrantV1(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfInstanceId=lcmocc.vnfInstanceId,
            vnfLcmOpOccId=lcmocc.id
        )

        return grant_req, grant

    def _create_grant_req_and_grant(self, lcmocc):
        grant_req, grant = self._make_grant_req_and_grant(lcmocc)
        grant_req.create(self.context)
        grant.create(self.context)
        lcmocc.grantId = grant.id
        lcmocc.update(self.context)

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'grant')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'post_grant')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'process')
    def test_start_lcm_op_completed(self, mocked_process, mocked_post_grant,
            mocked_grant, mocked_get_vnfd, mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc()
        mocked_get_vnfd.return_value = mock.Mock()
        grant_req, grant = self._make_grant_req_and_grant(lcmocc)
        lcmocc.grantId = grant.id
        mocked_grant.return_value = grant_req, grant

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run start_lcm_op
        self.conductor.start_lcm_op(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(3, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.STARTING, op_state[0])
        self.assertEqual(fields.LcmOperationStateType.PROCESSING, op_state[1])
        self.assertEqual(fields.LcmOperationStateType.COMPLETED, op_state[2])

        # check grant_req and grant are deleted
        self.assertRaises(sol_ex.GrantRequestOrGrantNotFound,
            lcmocc_utils.get_grant_req_and_grant, self.context, lcmocc)

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'grant')
    def test_start_lcm_op_rolled_back(self,
            mocked_grant, mocked_get_vnfd, mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc()
        mocked_get_vnfd.return_value = mock.Mock()
        ex = sol_ex.LocalNfvoGrantFailed(sol_detail="unit test")
        mocked_grant.side_effect = ex

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run start_lcm_op
        self.conductor.start_lcm_op(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(2, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.STARTING, op_state[0])
        self.assertEqual(fields.LcmOperationStateType.ROLLED_BACK, op_state[1])

        # check lcmocc.error
        # get lcmocc from DB to be sure lcmocc saved to DB
        lcmocc = lcmocc_utils.get_lcmocc(self.context, lcmocc.id)
        expected = ex.make_problem_details()
        self.assertEqual(expected, lcmocc.error.to_dict())

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'grant')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'post_grant')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'process')
    def test_start_lcm_op_failed_temp(self, mocked_process, mocked_post_grant,
            mocked_grant, mocked_get_vnfd, mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc()
        mocked_get_vnfd.return_value = mock.Mock()
        mocked_grant.return_value = self._make_grant_req_and_grant(lcmocc)
        ex = sol_ex.StackOperationFailed(sol_detail="unit test",
                                         sol_title="stack failed")
        mocked_process.side_effect = ex

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run start_lcm_op
        self.conductor.start_lcm_op(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(3, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.STARTING, op_state[0])
        self.assertEqual(fields.LcmOperationStateType.PROCESSING, op_state[1])
        self.assertEqual(fields.LcmOperationStateType.FAILED_TEMP, op_state[2])

        # check lcmocc.error
        # get lcmocc from DB to be sure lcmocc saved to DB
        lcmocc = lcmocc_utils.get_lcmocc(self.context, lcmocc.id)
        expected = ex.make_problem_details()
        self.assertEqual(expected, lcmocc.error.to_dict())

        # check grant_req and grant are saved to DB
        # it's OK if no exception raised
        lcmocc_utils.get_grant_req_and_grant(self.context, lcmocc)

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'post_grant')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'process')
    def test_retry_lcm_op_completed(self, mocked_process, mocked_post_grant,
            mocked_get_vnfd, mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc(
            op_state=fields.LcmOperationStateType.FAILED_TEMP)
        self._create_grant_req_and_grant(lcmocc)
        mocked_get_vnfd.return_value = mock.Mock()

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run retry_lcm_op
        self.conductor.retry_lcm_op(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(2, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.PROCESSING, op_state[0])
        self.assertEqual(fields.LcmOperationStateType.COMPLETED, op_state[1])

        # check grant_req and grant are deleted
        self.assertRaises(sol_ex.GrantRequestOrGrantNotFound,
            lcmocc_utils.get_grant_req_and_grant, self.context, lcmocc)

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'post_grant')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'process')
    def test_retry_lcm_op_failed_temp(self, mocked_process, mocked_post_grant,
            mocked_get_vnfd, mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc(
            op_state=fields.LcmOperationStateType.FAILED_TEMP)
        self._create_grant_req_and_grant(lcmocc)
        mocked_get_vnfd.return_value = mock.Mock()
        ex = sol_ex.StackOperationFailed(sol_detail="unit test",
                                         sol_title="stack failed")
        mocked_process.side_effect = ex

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run retry_lcm_op
        self.conductor.retry_lcm_op(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(2, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.PROCESSING, op_state[0])
        self.assertEqual(fields.LcmOperationStateType.FAILED_TEMP, op_state[1])

        # check lcmocc.error
        # get lcmocc from DB to be sure lcmocc saved to DB
        lcmocc = lcmocc_utils.get_lcmocc(self.context, lcmocc.id)
        expected = ex.make_problem_details()
        self.assertEqual(expected, lcmocc.error.to_dict())

        # check grant_req and grant remain
        # it's OK if no exception raised
        lcmocc_utils.get_grant_req_and_grant(self.context, lcmocc)

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'post_grant')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'rollback')
    def test_rollback_lcm_op_rolled_back(self, mocked_rollback,
            mocked_post_grant, mocked_get_vnfd,
            mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc(
            op_state=fields.LcmOperationStateType.FAILED_TEMP)
        self._create_grant_req_and_grant(lcmocc)
        mocked_get_vnfd.return_value = mock.Mock()

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run rollback_lcm_op
        self.conductor.rollback_lcm_op(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(2, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.ROLLING_BACK,
                         op_state[0])
        self.assertEqual(fields.LcmOperationStateType.ROLLED_BACK, op_state[1])

        # check grant_req and grant are deleted
        self.assertRaises(sol_ex.GrantRequestOrGrantNotFound,
            lcmocc_utils.get_grant_req_and_grant, self.context, lcmocc)

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'post_grant')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'rollback')
    def test_rollback_lcm_op_failed_temp(self, mocked_rollback,
            mocked_post_grant, mocked_get_vnfd,
            mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc(
            op_state=fields.LcmOperationStateType.FAILED_TEMP)
        self._create_grant_req_and_grant(lcmocc)
        mocked_get_vnfd.return_value = mock.Mock()
        ex = sol_ex.StackOperationFailed(sol_detail="unit test",
                                         sol_title="stack failed")
        mocked_rollback.side_effect = ex

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run rollback_lcm_op
        self.conductor.rollback_lcm_op(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(2, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.ROLLING_BACK,
                         op_state[0])
        self.assertEqual(fields.LcmOperationStateType.FAILED_TEMP, op_state[1])

        # check lcmocc.error
        # get lcmocc from DB to be sure lcmocc saved to DB
        lcmocc = lcmocc_utils.get_lcmocc(self.context, lcmocc.id)
        expected = ex.make_problem_details()
        self.assertEqual(expected, lcmocc.error.to_dict())

        # check grant_req and grant remain
        # it's OK if no exception raised
        lcmocc_utils.get_grant_req_and_grant(self.context, lcmocc)

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'process')
    def test_modify_vnfinfo_lcm_op_completed(self, mocked_process,
            mocked_get_vnfd, mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc(
            op_state=fields.LcmOperationStateType.PROCESSING)
        mocked_get_vnfd.return_value = mock.Mock()

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run modify_vnfinfo
        self.conductor.modify_vnfinfo(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(2, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.PROCESSING, op_state[0])
        self.assertEqual(fields.LcmOperationStateType.COMPLETED, op_state[1])

    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    @mock.patch.object(vnflcm_driver_v2.VnfLcmDriverV2, 'process')
    def test_modify_vnfinfo_lcm_op_failed_temp(self, mocked_process,
            mocked_get_vnfd, mocked_send_lcmocc_notification):
        # prepare
        lcmocc = self._create_inst_and_lcmocc(
            op_state=fields.LcmOperationStateType.PROCESSING)
        mocked_get_vnfd.return_value = mock.Mock()
        ex = sol_ex.StackOperationFailed(sol_detail="unit test",
                                         sol_title="stack failed")
        mocked_process.side_effect = ex

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run modify_vnfinfo
        self.conductor.modify_vnfinfo(self.context, lcmocc.id)

        # check operationState transition
        self.assertEqual(2, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(fields.LcmOperationStateType.PROCESSING, op_state[0])
        self.assertEqual(fields.LcmOperationStateType.FAILED_TEMP, op_state[1])

        # check lcmocc.error
        # get lcmocc from DB to be sure lcmocc saved to DB
        lcmocc = lcmocc_utils.get_lcmocc(self.context, lcmocc.id)
        expected = ex.make_problem_details()
        self.assertEqual(expected, lcmocc.error.to_dict())

    def _prepare_change_lcm_op_state(self, op_state):
        inst = objects.VnfInstanceV2(
            # required fields
            id=uuidutils.generate_uuid(),
            vnfdId=uuidutils.generate_uuid(),
            vnfProvider='provider',
            vnfProductName='product name',
            vnfSoftwareVersion='software version',
            vnfdVersion='vnfd version',
            instantiationState='INSTANTIATED'
        )

        req = {"flavourId": "simple"}  # instantiate request
        lcmocc = objects.VnfLcmOpOccV2(
            # required fields
            id=uuidutils.generate_uuid(),
            operationState=op_state,
            stateEnteredTime=datetime.utcnow(),
            startTime=datetime.utcnow(),
            vnfInstanceId=inst.id,
            operation=fields.LcmOperationType.SCALE,
            isAutomaticInvocation=False,
            isCancelPending=False,
            operationParams=req)

        inst.create(self.context)
        lcmocc.create(self.context)

        return lcmocc

    @ddt.data({'before_state': fields.LcmOperationStateType.STARTING,
               'after_state': fields.LcmOperationStateType.ROLLED_BACK},
              {'before_state': fields.LcmOperationStateType.PROCESSING,
               'after_state': fields.LcmOperationStateType.FAILED_TEMP},
              {'before_state': fields.LcmOperationStateType.ROLLING_BACK,
               'after_state': fields.LcmOperationStateType.FAILED_TEMP})
    @ddt.unpack
    @mock.patch.object(nfvo_client.NfvoClient, 'send_lcmocc_notification')
    @mock.patch.object(nfvo_client.NfvoClient, 'get_vnfd')
    def test_change_lcm_op_state(self, mocked_get_vnfd,
            mocked_send_lcmocc_notification, before_state, after_state):
        # prepare
        lcmocc = self._prepare_change_lcm_op_state(before_state)
        mocked_get_vnfd.return_value = mock.Mock()
        ex = sol_ex.ConductorProcessingError()

        op_state = []

        def _store_state(context, lcmocc, inst, endpoint):
            op_state.append(lcmocc.operationState)

        mocked_send_lcmocc_notification.side_effect = _store_state

        # run _change_lcm_op_state
        self.conductor._change_lcm_op_state()

        # check operationState transition
        self.assertEqual(1, mocked_send_lcmocc_notification.call_count)
        self.assertEqual(after_state, op_state[0])

        # check lcmocc.error
        # get lcmocc from DB to be sure lcmocc saved to DB
        lcmocc = lcmocc_utils.get_lcmocc(self.context, lcmocc.id)
        expected = ex.make_problem_details()
        self.assertEqual(expected, lcmocc.error.to_dict())



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add onboarded status for vnffgd and nsd

Revision ID: e9a1e47fb0b5
Revises: f5c1c3b0f6b4
Create Date: 2017-07-17 10:02:37.572587

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'e9a1e47fb0b5'
down_revision = 'f5c1c3b0f6b4'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.add_column('vnffgtemplates',
                  sa.Column('template_source',
                            sa.String(length=255),
                            server_default='onboarded'))
    op.execute("UPDATE vnffgtemplates set template_source='onboarded'"
        " WHERE template_source is NULL")

    op.add_column('nsd',
                  sa.Column('template_source',
                            sa.String(length=255),
                            server_default='onboarded'))
    op.execute("UPDATE nsd set template_source='onboarded'"
        " WHERE template_source is NULL")



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.tests.functional.sol_kubernetes.vnflcm import base as vnflcm_base


class VnfLcmKubernetesContainerUpdate(vnflcm_base.BaseVnfLcmKubernetesTest):

    @classmethod
    def setUpClass(cls):
        super(VnfLcmKubernetesContainerUpdate, cls).setUpClass()
        mgmt_rela_path = ("../../../../../samples/mgmt_driver/kubernetes/"
                          "container_update/container_update_mgmt.py")
        vnf_package_id_before, cls.vnfd_id_before = (
            cls._create_and_upload_vnf_package_add_mgmt(
                cls, cls.tacker_client, "test_cnf_container_update_before",
                {"key": "sample_container_update_before_functional"},
                mgmt_rela_path))
        cls.vnf_package_ids.append(vnf_package_id_before)

        vnf_package_id_after, cls.vnfd_id_after = (
            cls._create_and_upload_vnf_package_add_mgmt(
                cls, cls.tacker_client, "test_cnf_container_update_after",
                {"key": "sample_container_update_after_functional"},
                mgmt_rela_path))
        cls.vnf_package_ids.append(vnf_package_id_after)

    @classmethod
    def tearDownClass(cls):
        super(VnfLcmKubernetesContainerUpdate, cls).tearDownClass()

    def test_container_update_multi_kinds(self):
        vnf_instance_name = "container_update_multi_kinds"
        vnf_instance_description = "container update multi kinds"
        files = ["Files/kubernetes/configmap_1.yaml",
                 "Files/kubernetes/deployment.yaml",
                 "Files/kubernetes/pod_env.yaml",
                 "Files/kubernetes/pod_volume.yaml",
                 "Files/kubernetes/replicaset.yaml",
                 "Files/kubernetes/secret_1.yaml"]
        additional_param = {
            "lcm-kubernetes-def-files": files,
            "namespace": "default"}

        # instantiate
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id_before, "simple", vnf_instance_name,
            vnf_instance_description, additional_param)

        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        self.assertEqual(4, len(before_vnfc_rscs))

        # modify
        vnf_instance_name = "modify_vnf_after"
        configmap_secret_paths = [
            "Files/kubernetes/configmap_2.yaml",
            "Files/kubernetes/secret_2.yaml"]
        metadata = {"configmap_secret_paths": configmap_secret_paths}
        modify_request_body = {
            "vnfdId": self.vnfd_id_after,
            "vnfInstanceName": vnf_instance_name,
            "metadata": metadata
        }

        vnf_instance_after, after_vnfc_rscs = self._modify_vnf_instance(
            vnf_instance['id'], modify_request_body)

        self.assertEqual(4, len(after_vnfc_rscs))

        for after_vnfc_rsc in after_vnfc_rscs:
            for before_vnfc_rsc in before_vnfc_rscs:
                after_resource = after_vnfc_rsc['computeResource']
                before_resource = before_vnfc_rsc['computeResource']
                if after_vnfc_rsc['id'] == before_vnfc_rsc['id']:
                    if after_resource['vimLevelResourceType'] == 'Deployment':
                        # check stored pod name is changed (Deployment)
                        self.assertNotEqual(before_resource['resourceId'],
                                            after_resource['resourceId'])
                    else:
                        # check stored pod name is not changed (other)
                        self.assertEqual(before_resource['resourceId'],
                                         after_resource['resourceId'])

        self.assertEqual(
            self.vnfd_id_after, vnf_instance_after['vnfdId'])
        self.assertEqual(
            vnf_instance_name, vnf_instance_after['vnfInstanceName'])

        # terminate
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])



#    Copyright 2020 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.common import exceptions
from tacker import context
from tacker.objects import vnf_artifact
from tacker.objects import vnf_package
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestVnfPackageArtifact(SqlTestCase):

    def setUp(self):
        super(TestVnfPackageArtifact, self).setUp()
        self.context = context.get_admin_context()
        self.vnf_artifacts = self._create_vnf_artifact()

    def test_vnf_package_artifact_create(self):
        vnf_pack = vnf_package.VnfPackage(context=self.context,
                                          **fakes.vnf_package_data)
        vnf_pack.create()
        vnf_pack_artifact_data = fakes.vnf_pack_artifact_data(vnf_pack.id)
        result = vnf_artifact._vnf_artifacts_create(
            self.context, vnf_pack_artifact_data)
        self.assertTrue(result.id)

    def _create_vnf_artifact(self):
        vnf_pack = vnf_package.VnfPackage(context=self.context,
                                          **fakes.vnf_package_data)
        vnf_pack.create()
        vnf_pack_artifact_data = fakes.vnf_pack_artifact_data(vnf_pack.id)
        vnf_artifact_obj = vnf_artifact.VnfPackageArtifactInfo(
            context=self.context, **vnf_pack_artifact_data)
        vnf_artifact_obj.create()
        self.assertEqual('scripts/install.sh', vnf_artifact_obj.artifact_path)
        return vnf_artifact_obj

    def test_get_by_id(self):
        vnf_artifacts = vnf_artifact.VnfPackageArtifactInfo.get_by_id(
            self.context, self.vnf_artifacts.id)
        self.compare_obj(self.vnf_artifacts, vnf_artifacts)

    def test_get_by_id_with_no_existing_id(self):
        self.assertRaises(
            exceptions.VnfArtifactNotFound,
            vnf_artifact.VnfPackageArtifactInfo.get_by_id, self.context,
            uuidsentinel.invalid_uuid)

    def test_attribute_with_valid_data(self):
        data = {'id': self.vnf_artifacts.id}
        vnf_artifact_obj = vnf_artifact.VnfPackageArtifactInfo(
            context=self.context, **data)
        vnf_artifact_obj.obj_load_attr('artifact_path')
        self.assertEqual('scripts/install.sh', vnf_artifact_obj.artifact_path)

    def test_invalid_attribute(self):
        self.assertRaises(exceptions.ObjectActionError,
                          self.vnf_artifacts.obj_load_attr, 'invalid')

    def test_obj_load_attr_without_context(self):
        data = {'id': self.vnf_artifacts.id}
        vnf_artifact_obj = vnf_artifact.VnfPackageArtifactInfo(**data)
        self.assertRaises(exceptions.OrphanedObjectError,
                          vnf_artifact_obj.obj_load_attr, 'artifact_path')

    def test_obj_load_attr_without_id_in_object(self):
        data = {'artifact_path': self.vnf_artifacts.artifact_path}
        vnf_artifact_obj = vnf_artifact.VnfPackageArtifactInfo(
            context=self.context, **data)
        self.assertRaises(exceptions.ObjectActionError,
                          vnf_artifact_obj.obj_load_attr, 'artifact_path')



# Copyright 2015 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import copy
import inspect
import random
import string
import threading
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils

from tacker._i18n import _
from tacker.common import driver_manager
from tacker.common import exceptions
from tacker import context as t_context
from tacker.plugins.common import constants
from tacker.vnfm import utils as vnfm_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
OPTS = [
    cfg.IntOpt('check_intvl',
               default=10,
               help=_("check interval for monitor")),
]
CONF.register_opts(OPTS, group='monitor')


def config_opts():
    return [('monitor', OPTS),
            ('tacker', VNFMonitor.OPTS),
            ('tacker', VNFAlarmMonitor.OPTS),
            ('tacker', VNFAppMonitor.OPTS)]


class VNFMonitor(object):
    """VNF Monitor."""

    _instance = None
    _hosting_vnfs = dict()   # vnf_id => dict of parameters
    _status_check_intvl = 0
    _lock = threading.RLock()

    OPTS = [
        cfg.ListOpt(
            'monitor_driver', default=['ping', 'http_ping'],
            help=_('Monitor driver to communicate with '
                   'Hosting VNF/logical service '
                   'instance tacker plugin will use')),
    ]
    cfg.CONF.register_opts(OPTS, 'tacker')

    def __new__(cls, boot_wait, check_intvl=None):
        if not cls._instance:
            cls._instance = super(VNFMonitor, cls).__new__(cls)
        return cls._instance

    def __init__(self, boot_wait, check_intvl=None):
        self._monitor_manager = driver_manager.DriverManager(
            'tacker.tacker.monitor.drivers',
            cfg.CONF.tacker.monitor_driver)

        self.boot_wait = boot_wait
        if check_intvl is None:
            check_intvl = cfg.CONF.monitor.check_intvl
        self._status_check_intvl = check_intvl
        LOG.debug('Spawning VNF monitor thread')
        threading.Thread(target=self.__run__).start()

    def __run__(self):
        while(1):
            time.sleep(self._status_check_intvl)

            with self._lock:
                for hosting_vnf in VNFMonitor._hosting_vnfs.values():
                    if hosting_vnf.get('dead', False) or (
                            hosting_vnf['vnf']['status'] ==
                            constants.PENDING_HEAL):
                        LOG.debug(
                            'monitor skips for DEAD/PENDING_HEAL vnf %s',
                            hosting_vnf)
                        continue
                    try:
                        self.run_monitor(hosting_vnf)
                    except Exception as ex:
                        LOG.exception("Unknown exception: Monitoring failed "
                                      "for VNF '%s' due to '%s' ",
                                      hosting_vnf['id'], ex)

    @staticmethod
    def to_hosting_vnf(vnf_dict, action_cb):
        return {
            'id': vnf_dict['id'],
            'mgmt_ip_addresses': jsonutils.loads(
                vnf_dict['mgmt_ip_address']),
            'action_cb': action_cb,
            'vnf': vnf_dict,
            'monitoring_policy': jsonutils.loads(
                vnf_dict['attributes']['monitoring_policy'])
        }

    def add_hosting_vnf(self, new_vnf):
        LOG.debug('Adding host %(id)s, Mgmt IP %(ips)s',
                  {'id': new_vnf['id'],
                   'ips': new_vnf['mgmt_ip_addresses']})
        new_vnf['boot_at'] = timeutils.utcnow()
        with self._lock:
            VNFMonitor._hosting_vnfs[new_vnf['id']] = new_vnf

        attrib_dict = new_vnf['vnf']['attributes']
        mon_policy_dict = attrib_dict['monitoring_policy']
        evt_details = (("VNF added for monitoring. "
                        "mon_policy_dict = %s,") % (mon_policy_dict))
        vnfm_utils.log_events(t_context.get_admin_context(),
                              new_vnf['vnf'],
                              constants.RES_EVT_MONITOR, evt_details)

    def delete_hosting_vnf(self, vnf_id):
        LOG.debug('deleting vnf_id %(vnf_id)s', {'vnf_id': vnf_id})
        with self._lock:
            hosting_vnf = VNFMonitor._hosting_vnfs.pop(vnf_id, None)
            if hosting_vnf:
                LOG.debug('deleting vnf_id %(vnf_id)s, Mgmt IP %(ips)s',
                          {'vnf_id': vnf_id,
                           'ips': hosting_vnf['mgmt_ip_addresses']})

    def update_hosting_vnf(self, updated_vnf_dict, evt_details=None):
        with self._lock:
            vnf_to_update = VNFMonitor._hosting_vnfs.get(
                updated_vnf_dict.get('id'))
            if vnf_to_update:
                updated_vnf = copy.deepcopy(updated_vnf_dict)
                vnf_to_update['vnf'] = updated_vnf
                vnf_to_update['mgmt_ip_addresses'] = jsonutils.loads(
                    updated_vnf_dict['mgmt_ip_address'])

                if evt_details is not None:
                    vnfm_utils.log_events(t_context.get_admin_context(),
                                          vnf_to_update['vnf'],
                                          constants.RES_EVT_HEAL,
                                          evt_details=evt_details)

    def run_monitor(self, hosting_vnf):
        mgmt_ips = hosting_vnf['mgmt_ip_addresses']
        vdupolicies = hosting_vnf['monitoring_policy']['vdus']

        vnf_delay = hosting_vnf['monitoring_policy'].get(
            'monitoring_delay', self.boot_wait)

        for vdu in vdupolicies:
            if hosting_vnf.get('dead') or (
                    hosting_vnf['vnf']['status']) == constants.PENDING_HEAL:
                return

            policy = vdupolicies[vdu]
            for driver in policy:
                params = policy[driver].get('monitoring_params', {})

                vdu_delay = params.get('monitoring_delay', vnf_delay)

                if not timeutils.is_older_than(hosting_vnf['boot_at'],
                                               vdu_delay):
                    continue

                actions = policy[driver].get('actions', {})
                params['mgmt_ip'] = mgmt_ips[vdu]

                driver_return = self.monitor_call(driver,
                                                  hosting_vnf['vnf'],
                                                  params)

                LOG.debug('driver_return %s', driver_return)

                if driver_return in actions:
                    action = actions[driver_return]
                    hosting_vnf['action_cb'](action, vdu_name=vdu)

    def mark_dead(self, vnf_id):
        VNFMonitor._hosting_vnfs[vnf_id]['dead'] = True

    def _invoke(self, driver, **kwargs):
        method = inspect.stack()[1][3]
        return self._monitor_manager.invoke(
            driver, method, **kwargs)

    def monitor_get_config(self, vnf_dict):
        return self._invoke(
            vnf_dict, monitor=self, vnf=vnf_dict)

    def monitor_url(self, vnf_dict):
        return self._invoke(
            vnf_dict, monitor=self, vnf=vnf_dict)

    def monitor_call(self, driver, vnf_dict, kwargs):
        return self._invoke(driver,
                            vnf=vnf_dict, kwargs=kwargs)


class VNFAppMonitor(object):
    """VNF App monitor"""
    OPTS = [
        cfg.ListOpt(
            'app_monitor_driver', default=['zabbix'],
            help=_('App monitoring driver to communicate with '
                   'Hosting VNF/logical service '
                   'instance tacker plugin will use')),
    ]
    cfg.CONF.register_opts(OPTS, 'tacker')

    def __init__(self):
        self._application_monitor_manager = driver_manager.DriverManager(
            'tacker.tacker.app_monitor.drivers',
            cfg.CONF.tacker.app_monitor_driver)

    def _create_app_monitoring_dict(self, dev_attrs, mgmt_ip_address):
        app_policy = 'app_monitoring_policy'
        appmonitoring_dict = ast.literal_eval(
            dev_attrs[app_policy].decode('utf-8'))
        vdulist = appmonitoring_dict['vdus'].keys()

        for vduname in vdulist:
            temp = ast.literal_eval(mgmt_ip_address)
            appmonitoring_dict['vdus'][vduname]['mgmt_ip'] = temp[vduname]
        return appmonitoring_dict

    def create_app_dict(self, context, vnf_dict):
        dev_attrs = vnf_dict['attributes']
        mgmt_ip_address = vnf_dict['mgmt_ip_address'].decode("utf-8")
        return self._create_app_monitoring_dict(dev_attrs, mgmt_ip_address)

    def _invoke(self, driver, **kwargs):
        method = inspect.stack()[1][3]
        return self._application_monitor_manager.\
            invoke(driver, method, **kwargs)

    def add_to_appmonitor(self, applicationvnfdict, vnf_dict):
        vdunode = applicationvnfdict['vdus'].keys()
        driver = applicationvnfdict['vdus'][vdunode[0]]['name']
        kwargs = applicationvnfdict
        return self._invoke(driver, vnf=vnf_dict, kwargs=kwargs)


class VNFAlarmMonitor(object):
    """VNF Alarm monitor"""
    OPTS = [
        cfg.ListOpt(
            'alarm_monitor_driver', default=['ceilometer'],
            help=_('Alarm monitoring driver to communicate with '
                   'Hosting VNF/logical service '
                   'instance tacker plugin will use')),
    ]
    cfg.CONF.register_opts(OPTS, 'tacker')

    # get alarm here
    def __init__(self):
        self._alarm_monitor_manager = driver_manager.DriverManager(
            'tacker.tacker.alarm_monitor.drivers',
            cfg.CONF.tacker.alarm_monitor_driver)

    def update_vnf_with_alarm(self, plugin, context, vnf, policy_dict):
        triggers = policy_dict['triggers']
        alarm_url = dict()
        for trigger_name, trigger_dict in triggers.items():
            params = dict()
            params['vnf_id'] = vnf['id']
            params['mon_policy_name'] = trigger_name
            driver = trigger_dict['event_type']['implementation']
            # TODO(Tung Doan) trigger_dict.get('actions') needs to be used
            policy_action = trigger_dict.get('action')
            if len(policy_action) == 0:
                vnfm_utils.log_events(t_context.get_admin_context(), vnf,
                                      constants.RES_EVT_MONITOR,
                                      "Alarm not set: policy action missing")
                return
            # Other backend policies with the construct (policy, action)
            # ex: (SP1, in), (SP1, out)

            def _refactor_backend_policy(bk_policy_name, bk_action_name):
                policy = '%(policy_name)s-%(action_name)s' % {
                    'policy_name': bk_policy_name,
                    'action_name': bk_action_name}
                return policy

            for index, policy_action_name in enumerate(policy_action):
                filters = {'name': policy_action_name}
                bkend_policies =\
                    plugin.get_vnf_policies(context, vnf['id'], filters)
                if bkend_policies:
                    bkend_policy = bkend_policies[0]
                    if bkend_policy['type'] == constants.POLICY_SCALING:
                        cp = trigger_dict['condition'].\
                            get('comparison_operator')
                        scaling_type = 'out' if cp == 'gt' else 'in'
                        policy_action[index] = _refactor_backend_policy(
                            policy_action_name, scaling_type)

            # Support multiple action. Ex: respawn % notify
            action_name = '%'.join(policy_action)

            params['mon_policy_action'] = action_name
            alarm_url[trigger_name] =\
                self.call_alarm_url(driver, vnf, params)
            details = "Alarm URL set successfully: %s" % alarm_url
            vnfm_utils.log_events(t_context.get_admin_context(), vnf,
                                  constants.RES_EVT_MONITOR, details)
        return alarm_url

    def process_alarm_for_vnf(self, vnf, trigger):
        """call in plugin"""
        params = trigger['params']
        mon_prop = trigger['trigger']
        alarm_dict = dict()
        alarm_dict['alarm_id'] = params['data'].get('alarm_id')
        alarm_dict['status'] = params['data'].get('current')
        trigger_name, trigger_dict = list(mon_prop.items())[0]
        driver = trigger_dict['event_type']['implementation']
        return self.process_alarm(driver, vnf, alarm_dict)

    def _invoke(self, driver, **kwargs):
        method = inspect.stack()[1][3]
        return self._alarm_monitor_manager.invoke(
            driver, method, **kwargs)

    def call_alarm_url(self, driver, vnf_dict, kwargs):
        return self._invoke(driver,
                            vnf=vnf_dict, kwargs=kwargs)

    def process_alarm(self, driver, vnf_dict, kwargs):
        return self._invoke(driver,
                            vnf=vnf_dict, kwargs=kwargs)


class VNFReservationAlarmMonitor(VNFAlarmMonitor):
    """VNF Reservation Alarm monitor"""

    def update_vnf_with_reservation(self, plugin, context, vnf, policy_dict):

        alarm_url = dict()

        def create_alarm_action(action, action_list, scaling_type):
            params = dict()
            params['vnf_id'] = vnf['id']
            params['mon_policy_name'] = action
            driver = 'ceilometer'

            def _refactor_backend_policy(bk_policy_name, bk_action_name):
                policy = '%(policy_name)s%(action_name)s' % {
                    'policy_name': bk_policy_name,
                    'action_name': bk_action_name}
                return policy

            for index, policy_action_name in enumerate(action_list):
                filters = {'name': policy_action_name}
                bkend_policies = \
                    plugin.get_vnf_policies(context, vnf['id'], filters)
                if bkend_policies:
                    if constants.POLICY_SCALING in str(bkend_policies[0]):
                        action_list[index] = _refactor_backend_policy(
                            policy_action_name, scaling_type)

                # Support multiple action. Ex: respawn % notify
                action_name = '%'.join(action_list)
                params['mon_policy_action'] = action_name
                alarm_url[action] = \
                    self.call_alarm_url(driver, vnf, params)
                details = "Alarm URL set successfully: %s" % alarm_url
                vnfm_utils.log_events(t_context.get_admin_context(), vnf,
                                      constants.RES_EVT_MONITOR,
                                      details)

        before_end_action = policy_dict['reservation']['before_end_actions']
        end_action = policy_dict['reservation']['end_actions']
        start_action = policy_dict['reservation']['start_actions']

        scaling_policies = \
            plugin.get_vnf_policies(
                context, vnf['id'], filters={
                    'type': constants.POLICY_SCALING})

        if len(scaling_policies) == 0:
            raise exceptions.VnfPolicyNotFound(
                policy=constants.POLICY_SCALING, vnf_id=vnf['id'])

        for scaling_policy in scaling_policies:
            # validating start_action for scale-out policy action
            if scaling_policy['name'] not in start_action:
                raise exceptions.Invalid(
                    'Not a valid template: start_action must contain'
                    ' %s as scaling-out action' % scaling_policy['name'])

            # validating before_end and end_actions for scale-in policy action
            if scaling_policy['name'] not in before_end_action:
                if scaling_policy['name'] not in end_action:
                    raise exceptions.Invalid(
                        'Not a valid template:'
                        ' before_end_action or end_action'
                        ' should contain scaling policy: %s'
                        % scaling_policy['name'])

        for action in constants.RESERVATION_POLICY_ACTIONS:
            scaling_type = "-out" if action == 'start_actions' else "-in"
            create_alarm_action(action, policy_dict[
                'reservation'][action], scaling_type)

        return alarm_url

    def process_alarm_for_vnf(self, vnf, trigger):
        """call in plugin"""
        params = trigger['params']
        alarm_dict = dict()
        alarm_dict['alarm_id'] = params['data'].get('alarm_id')
        alarm_dict['status'] = params['data'].get('current')
        driver = 'ceilometer'
        return self.process_alarm(driver, vnf, alarm_dict)


class VNFMaintenanceAlarmMonitor(VNFAlarmMonitor):
    """VNF Maintenance Alarm monitor"""

    def update_vnf_with_maintenance(self, vnf, vdu_names):
        maintenance = dict()
        vdus = dict()
        params = dict()
        params['vnf_id'] = vnf['id']
        params['mon_policy_name'] = 'maintenance'
        params['mon_policy_action'] = vnf['tenant_id']
        driver = 'ceilometer'

        url = self.call_alarm_url(driver, vnf, params)
        maintenance['url'] = url[:url.rindex('/')]
        vdu_names.append('ALL')
        for vdu in vdu_names:
            access_key = ''.join(
                random.SystemRandom().choice(
                    string.ascii_lowercase + string.digits)
                for _ in range(8))
            vdus[vdu] = access_key
        maintenance.update({'vdus': vdus})
        details = "Alarm URL set successfully: %s" % maintenance['url']
        vnfm_utils.log_events(t_context.get_admin_context(), vnf,
                              constants.RES_EVT_MONITOR, details)
        return maintenance

    def process_alarm_for_vnf(self, vnf, trigger):
        """call in plugin"""
        params = trigger['params']
        alarm_dict = dict()
        alarm_dict['alarm_id'] = params['data'].get('alarm_id')
        alarm_dict['status'] = params['data'].get('current')
        driver = 'ceilometer'
        return self.process_alarm(driver, vnf, alarm_dict)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""rename device db tables

Revision ID: 22f5385a3d50
Revises: 22f5385a3d4f
Create Date: 2016-08-01 15:47:51.161749

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '22f5385a3d50'
down_revision = '22f5385a3d4f'

from alembic import op
import sqlalchemy as sa

from tacker.db import migration


def upgrade(active_plugins=None, options=None):
    # commands auto generated by Alembic - please adjust! #
    op.rename_table('devicetemplates', 'vnfd')
    op.rename_table('devicetemplateattributes', 'vnfd_attribute')
    op.rename_table('devices', 'vnf')
    op.rename_table('deviceattributes', 'vnf_attribute')
    migration.modify_foreign_keys_constraint_with_col_change(
        'vnfd_attribute', 'template_id', 'vnfd_id',
        sa.String(length=36))
    migration.modify_foreign_keys_constraint_with_col_change(
        'servicetypes', 'template_id', 'vnfd_id',
        sa.String(length=36))
    migration.modify_foreign_keys_constraint_with_col_change(
        'vnf', 'template_id', 'vnfd_id',
        sa.String(length=36))
    migration.modify_foreign_keys_constraint_with_col_change(
        'vnf_attribute', 'device_id', 'vnf_id',
        sa.String(length=36))
    # end Alembic commands #



# Copyright 2013, 2014 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.api import extensions


class VnfMGMTAbstractDriver(extensions.PluginInterface, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_type(self):
        """Return one of predefined type of the hosting vnf drivers."""
        pass

    @abc.abstractmethod
    def get_name(self):
        """Return a symbolic name for the service VM plugin."""
        pass

    @abc.abstractmethod
    def get_description(self):
        pass

    def mgmt_create_pre(self, plugin, context, vnf):
        pass

    def mgmt_create_post(self, plugin, context, vnf):
        pass

    def mgmt_update_pre(self, plugin, context, vnf):
        pass

    def mgmt_update_post(self, plugin, context, vnf):
        pass

    def mgmt_delete_pre(self, plugin, context, vnf):
        pass

    def mgmt_delete_post(self, plugin, context, vnf):
        pass

    def mgmt_get_config(self, plugin, context, vnf):
        """Get a dict of objects.

        Returns dict of file-like objects which will be passed to hosting
        vnf.
        It depends on drivers how to use it.
        for nova case, it can be used for meta data, file injection or
        config drive
        i.e.
        metadata case: nova --meta <key>=<value>
        file injection case: nova --file <dst-path>:<src-path>
        config drive case: nova --config-drive=true --file \
                                <dst-path>:<src-path>
        """
        return {}

    @abc.abstractmethod
    def mgmt_ip_address(self, plugin, context, vnf):
        pass

    @abc.abstractmethod
    def mgmt_call(self, plugin, context, vnf, kwargs):
        pass



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import shutil
import tempfile

from oslo_utils import uuidutils

from tacker.tests.functional.sol_v2 import paramgen
from tacker.tests.functional.sol_v2 import utils


zip_file_name = os.path.basename(os.path.abspath(".")) + '.zip'
tmp_dir = tempfile.mkdtemp()
vnfd_id = uuidutils.generate_uuid()
utils.make_zip(".", tmp_dir, vnfd_id)

shutil.copy(os.path.join(tmp_dir, zip_file_name), ".")
shutil.rmtree(tmp_dir)

create_req = paramgen.create_vnf_min(vnfd_id)
terminate_req = paramgen.terminate_vnf_min()
instantiate_req = paramgen.instantiate_vnf_min()
scaleout_req = paramgen.scaleout_vnf_min()
scalein_req = paramgen.scalein_vnf_min()
update_seq = paramgen.update_vnf_min()
# fake vnfc id, should be get from show vnf
VNFC_ID = "VDU1-9300a3cb-bd3b-45e4-9967-095040caf827"
heal_req = paramgen.heal_vnf_vnfc_min(VNFC_ID)
heal_without_parameter_req = paramgen.heal_vnf_all_min()
net_ids = {}
net_ids['ft-net1'] = '51e62f5f-3711-4182-b844-0f23e0408e51'
subnet_ids = {}
subnet_ids['ft-ipv4-subnet1'] = '8bf9b119-68bd-4e01-b518-dd4cde71687c'
subnet_ids['ft-ipv6-subnet1'] = '2bbaeb35-4d75-4aae-ab59-10c22a04d06b'
change_ext_conn_req = paramgen.change_ext_conn_min(net_ids, subnet_ids)

print('#####################################################################\n'
      '# vnfc id should be changed in heal req file by show vnf manually.  #\n'
      '#####################################################################')

with open("create_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(create_req, indent=2))

with open("terminate_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(terminate_req, indent=2))

with open("instantiate_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(instantiate_req, indent=2))

with open("scaleout_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(scaleout_req, indent=2))

with open("scalein_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(scalein_req, indent=2))

with open("update_seq", "w", encoding='utf-8') as f:
    f.write(json.dumps(update_seq, indent=2))

with open("heal_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_req, indent=2))

with open("heal_without_parameter_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(heal_without_parameter_req, indent=2))

with open("change_ext_conn_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(change_ext_conn_req, indent=2))



# Copyright 2013, 2014 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime

from oslo_db.exception import DBDuplicateEntry
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_utils import uuidutils

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc as orm_exc
from sqlalchemy import schema

from tacker._i18n import _
from tacker.api.v1 import attributes
from tacker.common import exceptions
import tacker.conf
from tacker import context as t_context
from tacker.db.common_services import common_services_db_plugin
from tacker.db import db_base
from tacker.db.db_sqlalchemy import models
from tacker.db import model_base
from tacker.db import models_v1
from tacker.db.nfvo import ns_db
from tacker.db import types
from tacker.extensions import vnfm
from tacker import manager
from tacker.plugins.common import constants

CONF = tacker.conf.CONF

LOG = logging.getLogger(__name__)
_ACTIVE_UPDATE = (constants.ACTIVE, constants.PENDING_UPDATE,
                  constants.PENDING_HEAL)
_ACTIVE_UPDATE_ERROR_DEAD = (
    constants.PENDING_CREATE, constants.ACTIVE, constants.PENDING_UPDATE,
    constants.PENDING_SCALE_IN, constants.PENDING_SCALE_OUT, constants.ERROR,
    constants.PENDING_DELETE, constants.DEAD, constants.PENDING_HEAL)
CREATE_STATES = (constants.PENDING_CREATE, constants.DEAD)


###########################################################################
# db tables

class VNFD(model_base.BASE, models_v1.HasId, models_v1.HasTenant,
           models_v1.Audit):
    """Represents VNFD to create VNF."""

    __tablename__ = 'vnfd'
    # Descriptive name
    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.Text)

    # service type that this service vm provides.
    # At first phase, this includes only single service
    # In future, single service VM may accommodate multiple services.
    service_types = orm.relationship('ServiceType', backref='vnfd')

    # driver to communicate with service management
    mgmt_driver = sa.Column(sa.String(255))

    # (key, value) pair to spin up
    attributes = orm.relationship('VNFDAttribute',
                                  backref='vnfd')

    # vnfd template source - inline or onboarded
    template_source = sa.Column(sa.String(255), server_default='onboarded')

    __table_args__ = (
        schema.UniqueConstraint(
            "tenant_id",
            "name",
            "deleted_at",
            name="uniq_vnfd0tenant_id0name0deleted_at"),
    )


class ServiceType(model_base.BASE, models_v1.HasId, models_v1.HasTenant):
    """Represents service type which hosting vnf provides.

    Since a vnf may provide many services, This is one-to-many
    relationship.
    """
    vnfd_id = sa.Column(types.Uuid, sa.ForeignKey('vnfd.id'),
                        nullable=False)
    service_type = sa.Column(sa.String(64), nullable=False)


class VNFDAttribute(model_base.BASE, models_v1.HasId):
    """Represents attributes necessary for spinning up VM in (key, value) pair

    key value pair is adopted for being agnostic to actuall manager of VMs.
    The interpretation is up to actual driver of hosting vnf.
    """

    __tablename__ = 'vnfd_attribute'
    vnfd_id = sa.Column(types.Uuid, sa.ForeignKey('vnfd.id'),
                        nullable=False)
    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.TEXT(65535), nullable=True)


class VNF(model_base.BASE, models_v1.HasId, models_v1.HasTenant,
          models_v1.Audit):
    """Represents vnfs that hosts services.

    Here the term, 'VM', is intentionally avoided because it can be
    VM or other container.
    """

    __tablename__ = 'vnf'
    vnfd_id = sa.Column(types.Uuid, sa.ForeignKey('vnfd.id'))
    vnfd = orm.relationship('VNFD')

    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.Text, nullable=True)

    # sufficient information to uniquely identify hosting vnf.
    # In case of openstack manager, it's UUID of heat stack.
    instance_id = sa.Column(sa.String(64), nullable=True)

    # For a management tool to talk to manage this hosting vnf.
    # opaque string.
    # e.g. (driver, mgmt_ip_address) = (ssh, ip address), ...
    mgmt_ip_address = sa.Column(sa.String(255), nullable=True)
    attributes = orm.relationship("VNFAttribute", backref="vnf")

    status = sa.Column(sa.String(64), nullable=False)
    vim_id = sa.Column(types.Uuid, sa.ForeignKey('vims.id'), nullable=False)
    placement_attr = sa.Column(types.Json, nullable=True)
    vim = orm.relationship('Vim')
    error_reason = sa.Column(sa.Text, nullable=True)

    __table_args__ = (
        schema.UniqueConstraint(
            "tenant_id",
            "name",
            "deleted_at",
            name="uniq_vnf0tenant_id0name0deleted_at"),
    )


class VNFAttribute(model_base.BASE, models_v1.HasId):
    """Represents kwargs necessary for spinning up VM in (key, value) pair.

    key value pair is adopted for being agnostic to actuall manager of VMs.
    The interpretation is up to actual driver of hosting vnf.
    """

    __tablename__ = 'vnf_attribute'
    vnf_id = sa.Column(types.Uuid, sa.ForeignKey('vnf.id'),
                       nullable=False)
    key = sa.Column(sa.String(255), nullable=False)
    # json encoded value. example
    # "nic": [{"net-id": <net-uuid>}, {"port-id": <port-uuid>}]
    value = sa.Column(sa.TEXT(65535), nullable=True)


class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):

    @property
    def _core_plugin(self):
        return manager.TackerManager.get_plugin()

    def subnet_id_to_network_id(self, context, subnet_id):
        subnet = self._core_plugin.get_subnet(context, subnet_id)
        return subnet['network_id']

    def __init__(self):
        super(VNFMPluginDb, self).__init__()
        self._cos_db_plg = common_services_db_plugin.CommonServicesPluginDb()

    def _get_resource(self, context, model, id):
        try:
            if uuidutils.is_uuid_like(id):
                return self._get_by_id(context, model, id)
            return self._get_by_name(context, model, id)
        except orm_exc.NoResultFound:
            if issubclass(model, VNFD):
                raise vnfm.VNFDNotFound(vnfd_id=id)
            elif issubclass(model, ServiceType):
                raise vnfm.ServiceTypeNotFound(service_type_id=id)
            if issubclass(model, VNF):
                raise vnfm.VNFNotFound(vnf_id=id)
            else:
                raise

    def _make_attributes_dict(self, attributes_db):
        return dict((attr.key, attr.value) for attr in attributes_db)

    def _make_service_types_list(self, service_types):
        return [service_type.service_type
                for service_type in service_types]

    def _make_vnfd_dict(self, vnfd, fields=None):
        res = {
            'attributes': self._make_attributes_dict(vnfd['attributes']),
            'service_types': self._make_service_types_list(
                vnfd.service_types)
        }
        key_list = ('id', 'tenant_id', 'name', 'description',
                    'mgmt_driver', 'created_at', 'updated_at',
                    'template_source')
        res.update((key, vnfd[key]) for key in key_list)
        return self._fields(res, fields)

    def _make_dev_attrs_dict(self, dev_attrs_db):
        return dict((arg.key, arg.value) for arg in dev_attrs_db)

    def _make_vnf_dict(self, vnf_db, fields=None):
        LOG.debug('vnf_db %s', vnf_db)
        LOG.debug('vnf_db attributes %s', vnf_db.attributes)
        res = {
            'vnfd':
            self._make_vnfd_dict(vnf_db.vnfd),
            'attributes': self._make_dev_attrs_dict(vnf_db.attributes),
        }
        key_list = ('id', 'tenant_id', 'name', 'description', 'instance_id',
                    'vim_id', 'placement_attr', 'vnfd_id', 'status',
                    'mgmt_ip_address', 'error_reason', 'created_at',
                    'updated_at')
        res.update((key, vnf_db[key]) for key in key_list)
        return self._fields(res, fields)

    @staticmethod
    def _mgmt_driver_name(vnf_dict):
        return vnf_dict['vnfd']['mgmt_driver']

    @staticmethod
    def _instance_id(vnf_dict):
        return vnf_dict['instance_id']

    def create_vnfd(self, context, vnfd):
        vnfd = vnfd['vnfd']
        LOG.debug('vnfd %s', vnfd)
        tenant_id = self._get_tenant_id_for_create(context, vnfd)
        service_types = vnfd.get('service_types')
        mgmt_driver = vnfd.get('mgmt_driver')
        template_source = vnfd.get("template_source")

        if (not attributes.is_attr_set(service_types)):
            LOG.debug('service types unspecified')
            raise vnfm.ServiceTypesNotSpecified()

        try:
            with context.session.begin(subtransactions=True):
                vnfd_id = uuidutils.generate_uuid()
                vnfd_db = VNFD(
                    id=vnfd_id,
                    tenant_id=tenant_id,
                    name=vnfd.get('name'),
                    description=vnfd.get('description'),
                    mgmt_driver=mgmt_driver,
                    template_source=template_source,
                    deleted_at=datetime.min)
                context.session.add(vnfd_db)
                for (key, value) in vnfd.get('attributes', {}).items():
                    attribute_db = VNFDAttribute(
                        id=uuidutils.generate_uuid(),
                        vnfd_id=vnfd_id,
                        key=key,
                        value=value)
                    context.session.add(attribute_db)
                for service_type in (item['service_type']
                                     for item in vnfd['service_types']):
                    service_type_db = ServiceType(
                        id=uuidutils.generate_uuid(),
                        tenant_id=tenant_id,
                        vnfd_id=vnfd_id,
                        service_type=service_type)
                    context.session.add(service_type_db)
        except DBDuplicateEntry as e:
            raise exceptions.DuplicateEntity(
                _type="vnfd",
                entry=e.columns)
        LOG.debug('vnfd_db %(vnfd_db)s %(attributes)s ',
                  {'vnfd_db': vnfd_db,
                   'attributes': vnfd_db.attributes})
        vnfd_dict = self._make_vnfd_dict(vnfd_db)
        LOG.debug('vnfd_dict %s', vnfd_dict)
        self._cos_db_plg.create_event(
            context, res_id=vnfd_dict['id'],
            res_type=constants.RES_TYPE_VNFD,
            res_state=constants.RES_EVT_ONBOARDED,
            evt_type=constants.RES_EVT_CREATE,
            tstamp=vnfd_dict[constants.RES_EVT_CREATED_FLD])
        return vnfd_dict

    def update_vnfd(self, context, vnfd_id,
                    vnfd):
        with context.session.begin(subtransactions=True):
            vnfd_db = self._get_resource(context, VNFD,
                                         vnfd_id)
            vnfd_db.update(vnfd['vnfd'])
            vnfd_db.update({'updated_at': timeutils.utcnow()})
            vnfd_dict = self._make_vnfd_dict(vnfd_db)
            self._cos_db_plg.create_event(
                context, res_id=vnfd_dict['id'],
                res_type=constants.RES_TYPE_VNFD,
                res_state=constants.RES_EVT_NA_STATE,
                evt_type=constants.RES_EVT_UPDATE,
                tstamp=vnfd_dict[constants.RES_EVT_UPDATED_FLD])
        return vnfd_dict

    def delete_vnfd(self,
                    context,
                    vnfd_id,
                    soft_delete=True):
        with context.session.begin(subtransactions=True):
            # TODO(yamahata): race. prevent from newly inserting hosting vnf
            #                 that refers to this vnfd
            vnfs_db = context.session.query(VNF).filter_by(
                vnfd_id=vnfd_id).first()
            if vnfs_db is not None and vnfs_db.deleted_at is None:
                raise vnfm.VNFDInUse(vnfd_id=vnfd_id)
            vnfd_db = self._get_resource(context, VNFD,
                                         vnfd_id)
            if soft_delete:
                vnfd_db.update({'deleted_at': timeutils.utcnow()})
                self._cos_db_plg.create_event(
                    context, res_id=vnfd_db['id'],
                    res_type=constants.RES_TYPE_VNFD,
                    res_state=constants.RES_EVT_NA_STATE,
                    evt_type=constants.RES_EVT_DELETE,
                    tstamp=vnfd_db[constants.RES_EVT_DELETED_FLD])
            else:
                context.session.query(ServiceType).filter_by(
                    vnfd_id=vnfd_id).delete()
                context.session.query(VNFDAttribute).filter_by(
                    vnfd_id=vnfd_id).delete()
                context.session.delete(vnfd_db)

    def get_vnfd(self, context, vnfd_id, fields=None):
        vnfd_db = self._get_resource(context, VNFD, vnfd_id)
        if not vnfd_db:
            raise exceptions.NotFound(resource='VNFD', name=vnfd_id)
        return self._make_vnfd_dict(vnfd_db)

    def get_vnfds(self, context, filters, fields=None):
        if ('template_source' in filters and
                filters['template_source'][0] == 'all'):
            filters.pop('template_source')
        return self._get_collection(context, VNFD,
                                    self._make_vnfd_dict,
                                    filters=filters, fields=fields)

    def choose_vnfd(self, context, service_type,
                    required_attributes=None):
        required_attributes = required_attributes or []
        LOG.debug('required_attributes %s', required_attributes)
        with context.session.begin(subtransactions=True):
            query = (
                context.session.query(VNFD).
                filter(
                    sa.exists().
                    where(sa.and_(
                        VNFD.id == ServiceType.vnfd_id,
                        ServiceType.service_type == service_type))))
            for key in required_attributes:
                query = query.filter(
                    sa.exists().
                    where(sa.and_(
                        VNFD.id ==
                        VNFDAttribute.vnfd_id,
                        VNFDAttribute.key == key)))
            LOG.debug('statements %s', query)
            vnfd_db = query.first()
            if vnfd_db:
                return self._make_vnfd_dict(vnfd_db)

    def _vnf_attribute_update_or_create(
            self, context, vnf_id, key, value):
        arg = (self._model_query(context, VNFAttribute).
               filter(VNFAttribute.vnf_id == vnf_id).
               filter(VNFAttribute.key == key).first())
        if arg:
            arg.value = value
        else:
            arg = VNFAttribute(
                id=uuidutils.generate_uuid(), vnf_id=vnf_id,
                key=key, value=value)
            context.session.add(arg)

    # called internally, not by REST API
    def _create_vnf_pre(self, context, vnf):
        LOG.debug('vnf %s', vnf)
        tenant_id = self._get_tenant_id_for_create(context, vnf)
        vnfd_id = vnf['vnfd_id']
        name = vnf.get('name')
        vnf_id = uuidutils.generate_uuid()
        attributes = vnf.get('attributes', {})
        vim_id = vnf.get('vim_id')
        placement_attr = vnf.get('placement_attr', {})
        try:
            with context.session.begin(subtransactions=True):
                vnfd_db = self._get_resource(context, VNFD,
                                             vnfd_id)
                vnf_db = VNF(id=vnf_id,
                             tenant_id=tenant_id,
                             name=name,
                             description=vnfd_db.description,
                             instance_id=None,
                             vnfd_id=vnfd_id,
                             vim_id=vim_id,
                             placement_attr=placement_attr,
                             status=constants.PENDING_CREATE,
                             error_reason=None,
                             deleted_at=datetime.min)
                context.session.add(vnf_db)
                for key, value in attributes.items():
                    arg = VNFAttribute(
                        id=uuidutils.generate_uuid(), vnf_id=vnf_id,
                        key=key, value=value)
                    context.session.add(arg)
        except DBDuplicateEntry as e:
            raise exceptions.DuplicateEntity(
                _type="vnf",
                entry=e.columns)
        evt_details = "VNF UUID assigned."
        self._cos_db_plg.create_event(
            context, res_id=vnf_id,
            res_type=constants.RES_TYPE_VNF,
            res_state=constants.PENDING_CREATE,
            evt_type=constants.RES_EVT_CREATE,
            tstamp=vnf_db[constants.RES_EVT_CREATED_FLD],
            details=evt_details)
        return self._make_vnf_dict(vnf_db)

    # called internally, not by REST API
    # intsance_id = None means error on creation
    def _create_vnf_post(self, context, vnf_id, instance_id,
                         mgmt_ip_address, vnf_dict):
        LOG.debug('vnf_dict %s', vnf_dict)
        with context.session.begin(subtransactions=True):
            query = (self._model_query(context, VNF).
                     filter(VNF.id == vnf_id).
                     filter(VNF.status.in_(CREATE_STATES)).
                     one())
            query.update({'instance_id': instance_id,
                          'mgmt_ip_address': mgmt_ip_address})
            if instance_id is None or vnf_dict['status'] == constants.ERROR:
                query.update({'status': constants.ERROR})

            for (key, value) in vnf_dict['attributes'].items():
                # do not store decrypted vim auth in vnf attr table
                if 'vim_auth' not in key:
                    self._vnf_attribute_update_or_create(context, vnf_id,
                                                         key, value)
        evt_details = ("Infra Instance ID created: %s and "
                       "Mgmt IP address set: %s") % (instance_id,
                                                     mgmt_ip_address)
        self._cos_db_plg.create_event(
            context, res_id=vnf_dict['id'],
            res_type=constants.RES_TYPE_VNF,
            res_state=vnf_dict['status'],
            evt_type=constants.RES_EVT_CREATE,
            tstamp=timeutils.utcnow(), details=evt_details)

    def _create_vnf_status(self, context, vnf_id, new_status):
        with context.session.begin(subtransactions=True):
            query = (self._model_query(context, VNF).
                     filter(VNF.id == vnf_id).
                     filter(VNF.status.in_(CREATE_STATES)).one())
            query.update({'status': new_status})
            self._cos_db_plg.create_event(
                context, res_id=vnf_id,
                res_type=constants.RES_TYPE_VNF,
                res_state=new_status,
                evt_type=constants.RES_EVT_CREATE,
                tstamp=timeutils.utcnow(), details="VNF creation completed")

    def _get_vnf_db(self, context, vnf_id, current_statuses):
        try:
            vnf_db = (
                self._model_query(context, VNF).
                filter(VNF.id == vnf_id).
                filter(VNF.status.in_(current_statuses)).
                with_for_update().one())
        except orm_exc.NoResultFound:
            raise vnfm.VNFNotFound(vnf_id=vnf_id)
        return vnf_db

    def _update_vnf_status_db(self, context, vnf_id, current_statuses,
                              new_status, vnf_db=None):
        if not vnf_db:
            vnf_db = self._get_vnf_db(context, vnf_id, current_statuses)
        if self.check_vnf_status_legality(vnf_db, vnf_id):
            vnf_db.update({'status': new_status})
        return vnf_db

    def _update_vnf_status_db_no_check(self, context, vnf_id, current_statuses,
                              new_status):
        vnf_db = self._get_vnf_db(context, vnf_id, current_statuses)
        vnf_db.update({'status': new_status})
        return vnf_db

    @staticmethod
    def check_vnf_status_legality(vnf_db, vnf_id):
        if vnf_db.status == constants.PENDING_DELETE:
            error_reason = _("Operation on PENDING_DELETE VNF "
                             "is not permitted. Please contact your "
                             "Administrator.")
            raise vnfm.VNFDeleteFailed(reason=error_reason)
        if(vnf_db.status in [constants.PENDING_UPDATE,
                             constants.PENDING_HEAL]):
            raise vnfm.VNFInUse(vnf_id=vnf_id)
        return True

    def update_vnf_cancel_status(self, context, vnf_id, status):
        with context.session.begin(subtransactions=True):
            self._update_vnf_status_db_no_check(
                context, vnf_id, [*constants.PENDING_STATUSES], status)

    def update_vnf_fail_status(self,
                               context,
                               vnf_id,
                               status):
        with context.session.begin(subtransactions=True):
            self._update_vnf_status_db(
                context, vnf_id, [constants.ERROR], status)

    def _update_vnf_scaling_status(self,
                                   context,
                                   policy,
                                   previous_statuses,
                                   status,
                                   mgmt_ip_address=None):
        with context.session.begin(subtransactions=True):
            vnf_db = self._update_vnf_status_db(
                context, policy['vnf']['id'], previous_statuses, status)
            if mgmt_ip_address:
                vnf_db.update({'mgmt_ip_address': mgmt_ip_address})
        updated_vnf_dict = self._make_vnf_dict(vnf_db)
        self._cos_db_plg.create_event(
            context, res_id=updated_vnf_dict['id'],
            res_type=constants.RES_TYPE_VNF,
            res_state=updated_vnf_dict['status'],
            evt_type=constants.RES_EVT_SCALE,
            tstamp=timeutils.utcnow())
        return updated_vnf_dict

    def _update_vnf_scaling_status_err(self,
                                       context,
                                       vnf_info):
        previous_statuses = ['PENDING_SCALE_OUT', 'PENDING_SCALE_IN', 'ACTIVE']

        try:
            with context.session.begin(subtransactions=True):
                self._update_vnf_status_db(
                    context, vnf_info['id'], previous_statuses, 'ERROR')
        except Exception as e:
            LOG.warning("Failed to revert scale info for vnf "
                        "instance %(id)s. Error: %(error)s",
                        {"id": vnf_info['id'], "error": e})
        self._cos_db_plg.create_event(
            context, res_id=vnf_info['id'],
            res_type=constants.RES_TYPE_VNF,
            res_state='ERROR',
            evt_type=constants.RES_EVT_SCALE,
            tstamp=timeutils.utcnow())

    def _update_vnf_scaling(self,
                            context,
                            vnf_info,
                            previous_statuses,
                            status):
        with context.session.begin(subtransactions=True):
            timestamp = timeutils.utcnow()
            (self._model_query(context, VNF).
             filter(VNF.id == vnf_info['id']).
             filter(VNF.status == previous_statuses).
             update({'status': status,
                     'updated_at': timestamp}))

            dev_attrs = vnf_info.get('attributes', {})
            (context.session.query(VNFAttribute).
             filter(VNFAttribute.vnf_id == vnf_info['id']).
             filter(~VNFAttribute.key.in_(dev_attrs.keys())).
             delete(synchronize_session='fetch'))

            for (key, value) in dev_attrs.items():
                if 'vim_auth' not in key:
                    self._vnf_attribute_update_or_create(
                        context, vnf_info['id'], key, value)
            self._cos_db_plg.create_event(
                context, res_id=vnf_info['id'],
                res_type=constants.RES_TYPE_VNF,
                res_state=status,
                evt_type=constants.RES_EVT_SCALE,
                tstamp=timestamp)

    def _update_vnf_pre(self, context, vnf_id, new_status):
        with context.session.begin(subtransactions=True):
            vnf_db = self._update_vnf_status_db(
                context, vnf_id, _ACTIVE_UPDATE, new_status)
        updated_vnf_dict = self._make_vnf_dict(vnf_db)
        if new_status in constants.VNF_STATUS_TO_EVT_TYPES:
            self._cos_db_plg.create_event(
                context, res_id=vnf_id,
                res_type=constants.RES_TYPE_VNF,
                res_state=updated_vnf_dict['status'],
                evt_type=constants.VNF_STATUS_TO_EVT_TYPES[new_status],
                tstamp=timeutils.utcnow())
        return updated_vnf_dict

    def _update_vnf_post(self, context, vnf_id, new_status,
                         new_vnf_dict, vnf_status, evt_type):
        updated_time_stamp = timeutils.utcnow()
        with context.session.begin(subtransactions=True):
            (self._model_query(context, VNF).
             filter(VNF.id == vnf_id).
             filter(VNF.status == vnf_status).
             update({'status': new_status,
                     'updated_at': updated_time_stamp,
                     'mgmt_ip_address': new_vnf_dict['mgmt_ip_address']}))

            dev_attrs = new_vnf_dict.get('attributes', {})
            (context.session.query(VNFAttribute).
             filter(VNFAttribute.vnf_id == vnf_id).
             filter(~VNFAttribute.key.in_(dev_attrs.keys())).
             delete(synchronize_session='fetch'))

            for (key, value) in dev_attrs.items():
                if 'vim_auth' not in key:
                    self._vnf_attribute_update_or_create(context, vnf_id,
                                                         key, value)
        self._cos_db_plg.create_event(
            context, res_id=vnf_id,
            res_type=constants.RES_TYPE_VNF,
            res_state=new_status,
            evt_type=evt_type,
            tstamp=updated_time_stamp)

    def _delete_vnf_pre(self, context, vnf_id, force_delete=False):
        with context.session.begin(subtransactions=True):

            nss_db = context.session.query(ns_db.NS).filter(
                ns_db.NS.vnf_ids.like("%" + vnf_id + "%")).first()

            if not force_delete:
                # If vnf is deleted by NFVO, then vnf_id would
                # exist in the nss_db otherwise it should be queried from
                # vnf db table.
                if nss_db is not None:
                    if nss_db.status not in [constants.PENDING_DELETE,
                                             constants.ERROR]:
                        raise vnfm.VNFInUse(vnf_id=vnf_id)
                else:
                    vnf_db = self._get_vnf_db(context, vnf_id,
                                              _ACTIVE_UPDATE_ERROR_DEAD)
                    if (vnf_db is not None and vnf_db.status == constants.
                            PENDING_CREATE):
                        raise vnfm.VNFInUse(
                            message="Operation on PENDING_CREATE VNF is not "
                                    "permitted.")

                    vnf_db = self._update_vnf_status_db(
                        context, vnf_id, _ACTIVE_UPDATE_ERROR_DEAD,
                        constants.PENDING_DELETE, vnf_db=vnf_db)
            else:
                vnf_db = self._update_vnf_status_db_no_check(context,
                    vnf_id, _ACTIVE_UPDATE_ERROR_DEAD,
                constants.PENDING_DELETE)
        deleted_vnf_db = self._make_vnf_dict(vnf_db)
        details = "VNF delete initiated" if not force_delete else \
            "VNF force delete initiated"
        self._cos_db_plg.create_event(
            context, res_id=vnf_id,
            res_type=constants.RES_TYPE_VNF,
            res_state=deleted_vnf_db['status'],
            evt_type=constants.RES_EVT_DELETE,
            tstamp=timeutils.utcnow(), details=details)
        return deleted_vnf_db

    def _delete_vnf_force(self, context, vnf_id):
        # Check mapping vnf in vnffg_db
        with context.session.begin(subtransactions=True):
            nss_db = context.session.query(ns_db.NS).filter(
                ns_db.NS.vnf_ids.like("%" + vnf_id + "%")).first()
            if nss_db:
                pass

    def _delete_vnf_post(self, context, vnf_dict, error,
                         soft_delete=True, force_delete=False):
        vnf_id = vnf_dict['id']
        with context.session.begin(subtransactions=True):
            if force_delete:
                query = (
                    self._model_query(context, VNF).
                    filter(VNF.id == vnf_id))
            else:
                query = (
                    self._model_query(context, VNF).
                    filter(VNF.id == vnf_id).
                    filter(VNF.status == constants.PENDING_DELETE))
            if error:
                query.update({'status': constants.ERROR})
                self._cos_db_plg.create_event(
                    context, res_id=vnf_id,
                    res_type=constants.RES_TYPE_VNF,
                    res_state=constants.ERROR,
                    evt_type=constants.RES_EVT_DELETE,
                    tstamp=timeutils.utcnow(),
                    details="VNF Delete ERROR")
            else:
                if soft_delete:
                    deleted_time_stamp = timeutils.utcnow()
                    query.update({'deleted_at': deleted_time_stamp})
                    self._cos_db_plg.create_event(
                        context, res_id=vnf_id,
                        res_type=constants.RES_TYPE_VNF,
                        res_state=constants.PENDING_DELETE,
                        evt_type=constants.RES_EVT_DELETE,
                        tstamp=deleted_time_stamp,
                        details="VNF Delete Complete")
                else:
                    (self._model_query(context, VNFAttribute).
                     filter(VNFAttribute.vnf_id == vnf_id).delete())
                    query.delete()

                # Delete corresponding vnfd
                if vnf_dict['vnfd']['template_source'] == "inline":
                    self.delete_vnfd(context, vnf_dict["vnfd_id"])

    # reference implementation. needs to be overrided by subclass
    def create_vnf(self, context, vnf):
        vnf_dict = self._create_vnf_pre(context, vnf)
        # start actual creation of hosting vnf.
        # Waiting for completion of creation should be done backgroundly
        # by another thread if it takes a while.
        instance_id = uuidutils.generate_uuid()
        vnf_dict['instance_id'] = instance_id
        self._create_vnf_post(context, vnf_dict['id'], instance_id, None,
                              vnf_dict)
        self._create_vnf_status(context, vnf_dict['id'],
                                constants.ACTIVE)
        return vnf_dict

    # reference implementation. needs to be overrided by subclass
    def update_vnf(self, context, vnf_id, vnf):
        new_status = constants.PENDING_UPDATE
        vnf_dict = self._update_vnf_pre(context, vnf_id, new_status)
        # start actual update of hosting vnf
        # waiting for completion of update should be done backgroundly
        # by another thread if it takes a while
        self._update_vnf_post(context, vnf_id,
                              constants.ACTIVE,
                              vnf_dict)
        return vnf_dict

    # reference implementation. needs to be overrided by subclass
    def delete_vnf(self, context, vnf_id, soft_delete=True):
        vnf_dict = self._delete_vnf_pre(context, vnf_id)
        # start actual deletion of hosting vnf.
        # Waiting for completion of deletion should be done backgroundly
        # by another thread if it takes a while.
        self._delete_vnf_post(context,
                              vnf_dict,
                              False,
                              soft_delete=soft_delete)

    def get_vnf(self, context, vnf_id, fields=None):
        vnf_db = self._get_resource(context, VNF, vnf_id)
        return self._make_vnf_dict(vnf_db, fields)

    def get_vnfs(self, context, filters=None, fields=None):
        return self._get_collection(context, VNF, self._make_vnf_dict,
                                    filters=filters, fields=fields)

    def set_vnf_error_status_reason(self, context, vnf_id, new_reason):
        with context.session.begin(subtransactions=True):
            (self._model_query(context, VNF).
                filter(VNF.id == vnf_id).
                update({'error_reason': new_reason}))

    def _mark_vnf_status(self, vnf_id, exclude_status, new_status):
        context = t_context.get_admin_context()
        with context.session.begin(subtransactions=True):
            try:
                vnf_db = (
                    self._model_query(context, VNF).
                    filter(VNF.id == vnf_id).
                    filter(~VNF.status.in_(exclude_status)).
                    with_for_update().one())
            except orm_exc.NoResultFound:
                LOG.warning('no vnf found %s', vnf_id)
                return False

            vnf_db.update({'status': new_status})
            self._cos_db_plg.create_event(
                context, res_id=vnf_id,
                res_type=constants.RES_TYPE_VNF,
                res_state=new_status,
                evt_type=constants.RES_EVT_MONITOR,
                tstamp=timeutils.utcnow())
        return True

    def _mark_vnf_error(self, vnf_id):
        return self._mark_vnf_status(
            vnf_id, [constants.DEAD], constants.ERROR)

    def _mark_vnf_dead(self, vnf_id):
        exclude_status = [
            constants.PENDING_CREATE,
            constants.PENDING_UPDATE,
            constants.PENDING_DELETE,
            constants.ERROR]
        return self._mark_vnf_status(
            vnf_id, exclude_status, constants.DEAD)

    def create_placement_constraint(self, context, placement_obj_list):
        context.session.add_all(placement_obj_list)

    def get_placement_constraint(self, context, vnf_instance_id):
        placement_constraint = (
            self._model_query(context, models.PlacementConstraint).filter(
                models.PlacementConstraint.vnf_instance_id == vnf_instance_id).
            filter(models.PlacementConstraint.deleted == 0).all())
        return placement_constraint

    def update_placement_constraint_heal(self, context,
                                         vnf_info,
                                         vnf_instance):
        if not vnf_info.get('grant'):
            return
        placement_obj_list = vnf_info['placement_obj_list']
        inst_info = vnf_instance.instantiated_vnf_info
        for vnfc in inst_info.vnfc_resource_info:
            for placement_obj in placement_obj_list:
                rsc_dict = jsonutils.loads(placement_obj.resource)
                for rsc in rsc_dict:
                    if vnfc.id == rsc.get('resource_id') and\
                       rsc.get('id_type') == 'GRANT':
                        rsc['id_type'] = 'RES_MGMT'
                        rsc['resource_id'] = vnfc.\
                            compute_resource.resource_id
                        rsc['vim_connection_id'] = vnfc.\
                            compute_resource.vim_connection_id
                placement_obj.resource = jsonutils.dumps(rsc_dict)
                self.update_placement_constraint(context, placement_obj)

    def delete_placement_constraint(self, context, vnf_instance_id):
        (self._model_query(context, models.PlacementConstraint).
            filter(
                models.PlacementConstraint.vnf_instance_id == vnf_instance_id).
            filter(models.PlacementConstraint.deleted == 0).
            update({'deleted': 0, 'deleted_at': timeutils.utcnow()}))

    def update_placement_constraint(self, context, placement_obj):
        (self._model_query(
            context,
            models.PlacementConstraint).filter(
                models.PlacementConstraint.id == placement_obj.id).
            filter(models.PlacementConstraint.deleted == 0).
            update({
                'resource': placement_obj.resource,
                'updated_at': timeutils.utcnow()}))

    def update_vnf_rollback_status_err(self,
                                       context,
                                       vnf_info):
        vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
        evt_type = constants.RES_EVT_CREATE
        if vnf_lcm_op_occs.operation == 'SCALE':
            evt_type = constants.RES_EVT_SCALE,
        self._cos_db_plg.create_event(context, res_id=vnf_info['id'],
                                      res_type=constants.RES_TYPE_VNF,
                                      res_state='ERROR', evt_type=evt_type,
                                      tstamp=timeutils.utcnow())

    def _update_vnf_rollback_pre(self,
                                 context,
                                 vnf_info):
        vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
        evt_type = constants.RES_EVT_CREATE
        if vnf_lcm_op_occs.operation == 'SCALE':
            evt_type = constants.RES_EVT_CREATE
        self._cos_db_plg.create_event(context, res_id=vnf_info['id'],
                                      res_type=constants.RES_TYPE_VNF,
                                      res_state='ROLL_BACK', evt_type=evt_type,
                                      tstamp=timeutils.utcnow())

    def _update_vnf_rollback(self,
                            context,
                            vnf_info,
                            previous_statuses,
                            status):
        with context.session.begin(subtransactions=True):
            timestamp = timeutils.utcnow()
            (self._model_query(context, VNF).
             filter(VNF.id == vnf_info['id']).
             filter(VNF.status == previous_statuses).
             update({'status': status,
                     'updated_at': timestamp}))

            dev_attrs = vnf_info.get('attributes', {})
            (context.session.query(VNFAttribute).
             filter(VNFAttribute.vnf_id == vnf_info['id']).
             filter(~VNFAttribute.key.in_(dev_attrs.keys())).
             delete(synchronize_session='fetch'))

            vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
            evt_type = constants.RES_EVT_CREATE
            if vnf_lcm_op_occs.operation == 'SCALE':
                for (key, value) in dev_attrs.items():
                    if 'vim_auth' not in key:
                        self._vnf_attribute_update_or_create(
                            context, vnf_info['id'], key, value)
                evt_type = constants.RES_EVT_SCALE
            self._cos_db_plg.create_event(context, res_id=vnf_info['id'],
                                          res_type=constants.RES_TYPE_VNF,
                                          res_state=status, evt_type=evt_type,
                                          tstamp=timestamp)



# Copyright 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Change nullable value of path_id

Revision ID: 13ecc2dd6f7f
Revises: 4747cc26b9c6
Create Date: 2018-07-24 16:47:01.378226

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '13ecc2dd6f7f'
down_revision = '4747cc26b9c6'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.alter_column('vnffgchains', 'path_id',
               existing_type=sa.String(length=255),
               nullable=True)
    op.alter_column('vnffgnfps', 'path_id',
               existing_type=sa.String(length=255),
               nullable=True)



# Copyright (C) 2018 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from webob import exc

from tacker.api.v1 import base as v1_base
from tacker.extensions import nfvo
from tacker.nfvo import nfvo_plugin
from tacker.tests.unit import base
from tacker import wsgi


def get_vim_config():
    return {
        "vim": {
            "tenant_id": 'test-project',
            "type": "openstack",
            "auth_url": 'http://localhost/identity',
            "auth_cred": {
                "username": "test_user",
                "user_domain_name": "Default",
                "password": "password"
            },
            "vim_project": {
                "name": "test_project",
                "project_domain_name": "Default"
            },
            "name": "VIM1",
            "description": "Additional site",
            "is_default": False
        }
    }


@ddt.ddt
class VIMCreateTestCase(base.TestCase):

    def setUp(self):
        super(VIMCreateTestCase, self).setUp()
        plugin = nfvo_plugin.NfvoPlugin()
        resource_name = 'vim'
        collection_name = resource_name + "s"
        attribute_info = nfvo.RESOURCE_ATTRIBUTE_MAP[collection_name]
        self.controller = v1_base.Controller(plugin, collection_name,
                                        resource_name, attribute_info)

    def _vim_create_response(self):
        return {
            'auth_cred': {
                'auth_url': 'http://localhost/identity',
                'cert_verify': 'False',
                'key_type': 'barbican_key',
                'password': '***',
                'project_domain_name': 'Default',
                'project_id': None,
                'project_name': 'nfv',
                'secret_uuid': '***',
                'user_domain_name': 'Default',
                'username': 'test_user'
            },
            'auth_url': 'http://localhost/identity',
            'created_at': None,
            'description': 'Additional site',
            'id': '73493efe-3616-414c-bf87-bf450d0b3650',
            'is_default': False,
            'name': 'VIM1',
            'placement_attr': {
                'regions': [
                    'RegionOne'
                ]
            },
            'shared': False,
            'status': 'PENDING',
            'tenant_id': 'test-project',
            'type': 'openstack',
            'updated_at': None,
            'vim_project': {
                'name': 'test_project'
            }
        }

    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'create_vim')
    def test_create_vim(self, mock_create_vim):
        vim_dict = get_vim_config()
        request = wsgi.Request.blank(
            "/vims.json", method='POST',
            headers={'Content-Type': "application/json"})
        request.environ['tacker.context'] = self.fake_admin_context()
        mock_create_vim.return_value = self._vim_create_response()

        result = self.controller.create(request, vim_dict)
        # End API response doesn't contain the 'shared' attribute so pop it
        # from dict
        resp_dict = self._vim_create_response()
        resp_dict.pop('shared')
        # Check whether VIM is created with the provided vim_details
        self.assertEqual(resp_dict, result['vim'])

    @ddt.data({'is_default': 'ABC'},
              {'is_default': 123},
              {'is_default': ''})
    def test_create_vim_with_invalid_is_default(self, value):
        vim_dict = get_vim_config()
        vim_dict['vim']['is_default'] = value
        request = wsgi.Request.blank("/vims.json", method='POST',
                  headers={'Content-Type': "application/json"})
        request.environ['tacker.context'] = self.fake_admin_context()
        msg = ("Invalid input for is_default. Reason: '%s' is not a "
               "valid boolean value." % vim_dict['vim']['is_default'])
        exp = self.assertRaises(exc.HTTPBadRequest,
                                self.controller.create,
                                request, vim_dict)
        self.assertEqual(msg, str(exp))

    @ddt.data("", " ", None, 123)
    def test_create_vim_with_invalid_type(self, value):
        vim_dict = get_vim_config()
        vim_dict['vim']['type'] = value
        request = wsgi.Request.blank("/vims.json", method='POST',
            headers={'Content-Type': "application/json"})
        request.environ['tacker.context'] = self.fake_admin_context()
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          request, vim_dict)

    @ddt.data('', 'testing', {})
    def test_create_vim_with_invalid_auth_cred(self, value):
        vim_dict = get_vim_config()
        vim_dict['vim']['auth_cred'] = value
        request = wsgi.Request.blank("/vims.json", method='POST',
                headers={'Content-Type': "application/json"})
        request.environ['tacker.context'] = self.fake_admin_context()
        msg = ("Invalid input for auth_cred. Reason: '%s' is "
               "not a valid dictionary or it is an empty"
               " dictionary.") % vim_dict['vim']['auth_cred']
        exp = self.assertRaises(exc.HTTPBadRequest,
                                self.controller.create,
                                request, vim_dict)
        self.assertEqual(msg, str(exp))

    @ddt.data('', 'testing', {})
    def test_create_vim_invalid_vim_project(self, value):
        vim_dict = get_vim_config()
        vim_dict['vim']['vim_project'] = value
        request = wsgi.Request.blank("/vims.json", method='POST',
                headers={'Content-Type': "application/json"})
        request.environ['tacker.context'] = self.fake_admin_context()
        msg = ("Invalid input for vim_project. Reason: '%s' is"
               " not a valid dictionary or it is an empty"
               " dictionary.") % vim_dict['vim']['vim_project']
        exp = self.assertRaises(exc.HTTPBadRequest,
                                self.controller.create,
                                request, vim_dict)
        self.assertEqual(msg, str(exp))



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""change_vim_shared_property_to_false

Revision ID: 31acbaeb8299
Revises: e7993093baf1
Create Date: 2017-05-30 23:46:20.034085

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '31acbaeb8299'
down_revision = 'e7993093baf1'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.alter_column('vims', 'shared',
                    existing_type=sa.Boolean(),
                    server_default=sa.text('false'),
                    nullable=False)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.tests.functional.sol_kubernetes.vnflcm import base as vnflcm_base


class VnfLcmKubernetesHelmTest(vnflcm_base.BaseVnfLcmKubernetesTest):

    @classmethod
    def setUpClass(cls):
        super(VnfLcmKubernetesHelmTest, cls).setUpClass()
        vnf_package_id, cls.vnfd_id = \
            cls._create_and_upload_vnf_package(
                cls, cls.tacker_client, "test_cnf_helmchart",
                {"key": "sample_helmchart_functional"})
        cls.vnf_package_ids.append(vnf_package_id)

    @classmethod
    def tearDownClass(cls):
        super(VnfLcmKubernetesHelmTest, cls).tearDownClass()

    def _test_scale_cnf(self, vnf_instance, aspect_id):
        """Test scale in/out CNF"""
        scale_level = self._get_scale_level_by_aspect_id(
            vnf_instance, aspect_id)

        # test scale out
        scale_level = self._test_scale(
            vnf_instance['id'], 'SCALE_OUT', aspect_id, scale_level)

        # test scale in
        scale_level = self._test_scale(
            vnf_instance['id'], 'SCALE_IN', aspect_id, scale_level)

    def _test_heal_cnf_with_sol002(self, vnf_instance):
        """Test heal as per SOL002 for CNF"""
        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)

        # get vnfc_instance_id of heal target
        before_pod_name = {}
        vnfc_instance_id = []
        for vnfc_rsc in before_vnfc_rscs:
            if vnfc_rsc['vduId'] == "vdu1":
                before_pod_name['vdu1'] = \
                    vnfc_rsc['computeResource']['resourceId']
            elif vnfc_rsc['vduId'] == "vdu2":
                before_pod_name['vdu2'] = \
                    vnfc_rsc['computeResource']['resourceId']
            vnfc_instance_id.append(vnfc_rsc['id'])

        # test heal SOL-002 (partial heal)
        after_vnfc_rscs = self._test_heal(vnf_instance, vnfc_instance_id)
        for vnfc_rsc in after_vnfc_rscs:
            after_pod_name = vnfc_rsc['computeResource']['resourceId']
            if vnfc_rsc['vduId'] == "vdu1":
                # check stored pod name is changed (vdu1)
                compute_resource = vnfc_rsc['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertNotEqual(after_pod_name, before_pod_name['vdu1'])
            elif vnfc_rsc['vduId'] == "vdu2":
                # check stored pod name is changed (vdu2)
                compute_resource = vnfc_rsc['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertNotEqual(after_pod_name, before_pod_name['vdu2'])

    def _test_heal_cnf_with_sol003(self, vnf_instance):
        """Test heal as per SOL003 for CNF"""
        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)

        # test heal SOL-003 (entire heal)
        vnfc_instance_id = []
        after_vnfc_rscs = self._test_heal(vnf_instance, vnfc_instance_id)
        # check id and pod name (as computeResource.resourceId) is changed
        for before_vnfc_rsc in before_vnfc_rscs:
            for after_vnfc_rsc in after_vnfc_rscs:
                self.assertNotEqual(
                    before_vnfc_rsc['id'], after_vnfc_rsc['id'])
                self.assertNotEqual(
                    before_vnfc_rsc['computeResource']['resourceId'],
                    after_vnfc_rsc['computeResource']['resourceId'])

    def test_vnflcm_with_helmchart(self):
        """Test LCM using Helm chart

        This test will instantiate, scale, heal, terminate cnf by using
        local and external Helm charts.
        """
        vnf_instance_name = "cnf_with_helmchart"
        vnf_instance_description = "cnf with helmchart"
        helmchartfile_path = "Files/kubernetes/localhelm-0.1.0.tgz"
        inst_additional_param = {
            "namespace": "default",
            "use_helm": "true",
            "using_helm_install_param": [
                {
                    "exthelmchart": "false",
                    "helmchartfile_path": helmchartfile_path,
                    "helmreleasename": "vdu1",
                    "helmparameter": [
                        "service.port=8081"
                    ]
                },
                {
                    "exthelmchart": "true",
                    "helmreleasename": "vdu2",
                    "helmrepositoryname": "bitnami",
                    "helmchartname": "apache",
                    "exthelmrepo_url": "https://charts.bitnami.com/bitnami"
                }
            ],
            "helm_replica_values": {
                "vdu1_aspect": "replicaCount",
                "vdu2_aspect": "replicaCount"
            }
        }
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "helmchart", vnf_instance_name,
            vnf_instance_description, inst_additional_param)

        self._test_scale_cnf(vnf_instance, aspect_id="vdu1_aspect")
        self._test_scale_cnf(vnf_instance, aspect_id="vdu2_aspect")
        self._test_heal_cnf_with_sol002(vnf_instance)
        self._test_heal_cnf_with_sol003(vnf_instance)

        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])



# Copyright 2011 Citrix Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob.dec

import oslo_i18n

from tacker._i18n import _
from tacker.api.views import versions as versions_view
from tacker import wsgi


class Versions(object):

    @classmethod
    def factory(cls, global_config, **local_config):
        return cls()

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        """Respond to a request for all Tacker API versions."""
        version_objs = [
            {
                "id": "v1.0",
                "status": "CURRENT",
            },
        ]

        if req.path != '/':
            language = req.best_match_language()
            msg = _('Unknown API version specified')
            msg = oslo_i18n.translate(msg, language)
            return webob.exc.HTTPNotFound(explanation=msg)

        builder = versions_view.get_view_builder(req)
        versions = [builder.build(version) for version in version_objs]
        response = dict(versions=versions)
        metadata = {}

        content_type = req.best_match_content_type()
        body = (wsgi.Serializer(metadata=metadata).
                serialize(response, content_type))

        response = webob.Response()
        response.content_type = content_type
        response.body = body

        return response



# Copyright (C) 2022 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from kubernetes import client
import os

from tacker import objects
from tacker.objects import fields
from tacker.tests import uuidsentinel


def get_vnf_instance_object(
        instantiation_state=fields.VnfInstanceState.NOT_INSTANTIATED):

    inst_vnf_info = get_vnf_instantiated_info()

    vnf_instance = objects.VnfInstance(
        id=uuidsentinel.vnf_instance_id,
        vnf_instance_name="Test-Vnf-Instance",
        vnf_instance_description="vnf instance description",
        instantiation_state=instantiation_state, vnfd_id=uuidsentinel.vnfd_id,
        vnf_provider="sample provider", vnf_product_name="vnf product name",
        vnf_software_version='1.0', vnfd_version="2",
        instantiated_vnf_info=inst_vnf_info,
        vnf_metadata={"namespace": "default"}
    )

    return vnf_instance


def get_vnf_instantiated_info(flavour_id='simple',
        instantiation_level_id=None, vnfc_resource_info=None,
        virtual_storage_resource_info=None,
        vnf_virtual_link_resource_info=None,
        ext_managed_virtual_link_info=None,
        ext_virtual_link_info=None,
        ext_cp_info=None):

    vnfc_resource_info = vnfc_resource_info or []
    vnf_virtual_link_resource_info = vnf_virtual_link_resource_info or []
    virtual_storage_resource_info = virtual_storage_resource_info or []
    ext_managed_virtual_link_info = ext_managed_virtual_link_info or []
    ext_virtual_link_info = ext_virtual_link_info or []
    ext_cp_info = ext_cp_info or []

    inst_vnf_info = objects.InstantiatedVnfInfo(
        flavour_id=flavour_id,
        instantiation_level_id=instantiation_level_id,
        instance_id='9693ab84-6da8-5926-5e7a-3a1b963156c0',
        vnfc_resource_info=vnfc_resource_info,
        vnf_virtual_link_resource_info=vnf_virtual_link_resource_info,
        virtual_storage_resource_info=virtual_storage_resource_info,
        ext_managed_virtual_link_info=ext_managed_virtual_link_info,
        ext_virtual_link_info=ext_virtual_link_info,
        ext_cp_info=ext_cp_info,
        additional_params=return_cnf_additional_params())

    return inst_vnf_info


def fake_pod():
    return client.V1Pod(
        api_version='v1',
        kind='Pod',
        metadata=client.V1ObjectMeta(
            name='curry-test001',
            namespace='curryns'
        ),
        spec=client.V1PodSpec(
            containers=[
                client.V1Container(
                    image="curry",
                    name="curry"
                )
            ]
        ),
        status=client.V1PodStatus(
            phase='Running',
        )
    )


def fake_list_pod():
    return client.V1PodList(
        items=[
            client.V1Pod(
                api_version='v1',
                kind='Pod',
                metadata=client.V1ObjectMeta(
                    name='curry-test001',
                    namespace='curryns'
                ),
                spec=client.V1PodSpec(
                    containers=[
                        client.V1Container(
                            image="curry",
                            name="curry"
                        )
                    ]
                ),
                status=client.V1PodStatus(
                    phase='Pending',
                )
            )
        ]
    )


def get_fake_pod_info(kind, name='fake_name', pod_status='Running',
                      pod_name=None):
    if not pod_name:
        if kind == 'Deployment':
            pod_name = _('{name}-1234567890-abcde').format(name=name)
        elif kind in ('ReplicaSet', 'DaemonSet'):
            pod_name = _('{name}-12345').format(name=name)
        elif kind == 'StatefulSet':
            pod_name = _('{name}-1').format(name=name)
        elif kind == 'Pod':
            pod_name = name
    return client.V1Pod(
        metadata=client.V1ObjectMeta(
            name=pod_name,
            creation_timestamp=datetime.datetime.now().isoformat('T')),
        status=client.V1PodStatus(phase=pod_status))


def vnfd_dict_cnf():
    tacker_dir = os.getcwd()
    def_dir = tacker_dir + "/samples/vnf_packages/Definitions/"
    vnfd_dict = {
        "tosca_definitions_version": "tosca_simple_yaml_1_2",
        "description": "Sample VNF flavour for Sample VNF",
        "imports": [
            def_dir + "etsi_nfv_sol001_common_types.yaml",
            def_dir + "etsi_nfv_sol001_vnfd_types.yaml",
            def_dir + "helloworld3_types.yaml"],
        "topology_template": {
            "node_templates": {
                "VNF": {
                    "type": "company.provider.VNF",
                    "properties": {
                        "flavour_description": "A simple flavour"}},
                "VDU1": {
                    "type": "tosca.nodes.nfv.Vdu.Compute",
                    "properties": {
                        "name": "vdu1",
                        "description": "vdu1 compute node",
                        "vdu_profile": {
                            "min_number_of_instances": 1,
                            "max_number_of_instances": 3}}}},
            "policies": [
                {
                    "scaling_aspects": {
                        "type": "tosca.policies.nfv.ScalingAspects",
                        "properties": {
                            "aspects": {
                                "vdu1_aspect": {
                                    "name": "vdu1_aspect",
                                    "description": "vdu1 scaling aspect",
                                    "max_scale_level": 2,
                                    "step_deltas": ["delta_1"]}}}}},
                {
                    "vdu1_initial_delta": {
                        "type": "tosca.policies.nfv.VduInitialDelta",
                        "properties": {
                            "initial_delta": {
                                "number_of_instances": 0}},
                        "targets": ["VDU1"]}},
                {
                    "vdu1_scaling_aspect_deltas": {
                        "type": "tosca.policies.nfv.VduScalingAspectDeltas",
                        "properties": {
                            "aspect": "vdu1_aspect",
                            "deltas": {
                                "delta_1": {
                                    "number_of_instances": 1}}},
                        "targets": ["VDU1"]}},
                {
                    "instantiation_levels": {
                        "type": "tosca.policies.nfv.InstantiationLevels",
                        "properties": {
                            "levels": {
                                "instantiation_level_1": {
                                    "description": "Smallest size",
                                    "scale_info": {
                                        "vdu1_aspect": {
                                            "scale_level": 0}}},
                                "instantiation_level_2": {
                                    "description": "Largest size",
                                    "scale_info": {
                                        "vdu1_aspect": {
                                            "scale_level": 2}}}
                            },
                            "default_level": "instantiation_level_1"}}},
                {
                    "vdu1_instantiation_levels": {
                        "type": "tosca.policies.nfv.VduInstantiationLevels",
                        "properties": {
                            "levels": {
                                "instantiation_level_1": {
                                    "number_of_instances": 0},
                                "instantiation_level_2": {
                                    "number_of_instances": 2}}},
                        "targets": ["VDU1"]}}
            ]
        }
    }
    return vnfd_dict


def return_cnf_additional_params():
    additional_params = {
        "lcm-kubernetes-def-files": [
            "Files/kubernetes/configmap_1.yaml",
            "Files/kubernetes/pod_volume.yaml",
            "Files/kubernetes/replicaset.yaml",
            "Files/kubernetes/secret_1.yaml"
        ],
        "namespace": "default"
    }
    return additional_params



# Copyright 2016 Red Hat Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils
from tacker.common import log
from tacker.nfvo.drivers.vnffg import abstract_vnffg_driver

LOG = logging.getLogger(__name__)


class VNFFGNoop(abstract_vnffg_driver.VnffgAbstractDriver):

    """Noop driver for VNFFG tests"""

    def __init__(self):
        super(VNFFGNoop, self).__init__()
        self._instances = set()

    def get_type(self):
        return 'noop'

    def get_name(self):
        return 'noop'

    def get_description(self):
        return 'VNFFG Noop driver'

    @log.log
    def create_chain(self, name, fc_id, vnfs, auth_attr=None):
        instance_id = uuidutils.generate_uuid()
        self._instances.add(instance_id)
        return instance_id

    @log.log
    def update_chain(self, chain_id, fc_ids, vnfs, auth_attr=None):
        if chain_id not in self._instances:
            LOG.debug('Chain not found')
            raise ValueError('No chain instance %s' % chain_id)

    @log.log
    def delete_chain(self, chain_id, auth_attr=None):
        self._instances.remove(chain_id)

    @log.log
    def create_flow_classifier(self, name, fc, auth_attr=None):
        instance_id = uuidutils.generate_uuid()
        self._instances.add(instance_id)
        return instance_id

    @log.log
    def update_flow_classifier(self, fc_id, fc, auth_attr=None):
        if fc_id not in self._instances:
            LOG.debug('FC not found')
            raise ValueError('No FC instance %s' % fc_id)

    @log.log
    def delete_flow_classifier(self, fc_id, auth_attr=None):
        self._instances.remove(fc_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import netaddr

from oslo_config import cfg
from oslo_log import log as logging

from tacker._i18n import _
from tacker.agent.linux import utils as linux_utils
from tacker.common import log
from tacker.vnfm.monitor_drivers import abstract_driver


LOG = logging.getLogger(__name__)
OPTS = [
    cfg.IntOpt('count', default=5,
               help=_('Number of ICMP packets to send')),
    cfg.FloatOpt('timeout', default=5,
               help=_('Number of seconds to wait for a response')),
    cfg.FloatOpt('interval', default=1,
               help=_('Number of seconds to wait between packets')),
    cfg.IntOpt('retry', default=1,
               help=_('Number of ping retries'))
]
cfg.CONF.register_opts(OPTS, 'monitor_ping')


def config_opts():
    return [('monitor_ping', OPTS)]


class VNFMonitorPing(abstract_driver.VNFMonitorAbstractDriver):
    def get_type(self):
        return 'ping'

    def get_name(self):
        return 'ping'

    def get_description(self):
        return 'Tacker VNFMonitor Ping Driver'

    def monitor_url(self, plugin, context, vnf):
        LOG.debug('monitor_url %s', vnf)
        return vnf.get('monitor_url', '')

    def _is_pingable(self, mgmt_ip="", count=None, timeout=None,
                     interval=None, retry=None, **kwargs):
        """Checks whether an IP address is reachable by pinging.

        Use linux utils to execute the ping (ICMP ECHO) command.
        Sends 5 packets with an interval of 1 seconds and timeout of 1
        seconds. Runtime error implies unreachability else IP is pingable.
        :param ip: IP to check
        :return: bool - True or string 'failure' depending on pingability.
        """
        cmd_ping = 'ping'
        if netaddr.valid_ipv6(mgmt_ip):
            cmd_ping = 'ping6'

        if not count:
            count = cfg.CONF.monitor_ping.count
        if not timeout:
            timeout = cfg.CONF.monitor_ping.timeout
        if not interval:
            interval = cfg.CONF.monitor_ping.interval
        if not retry:
            retry = cfg.CONF.monitor_ping.retry

        ping_cmd = [cmd_ping,
                    '-c', count,
                    '-W', timeout,
                    '-i', interval,
                    mgmt_ip]

        for retry_range in range(int(retry)):
            try:
                linux_utils.execute(ping_cmd, check_exit_code=True)
                return True
            except RuntimeError:
                LOG.warning("Cannot ping ip address: %s", mgmt_ip)
        return 'failure'

    @log.log
    def monitor_call(self, vnf, kwargs):
        if not kwargs['mgmt_ip']:
            return

        return self._is_pingable(**kwargs)



# Copyright 2015 Brocade Communications System, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import http.server
import os
import re
import shutil
import threading

from oslo_utils import uuidutils
import socketserver
import tempfile
import yaml
import zipfile


def read_file(input_file):
    yaml_file = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                             'etc/samples/' + str(input_file)))
    with open(yaml_file, 'r') as f:
        return f.read()


def _update_unique_id_in_yaml(data, uid):
    try:
        prop = data['topology_template']['node_templates']['VNF'][
            'properties']
        if (prop.get('descriptor_id', None)):
            prop['descriptor_id'] = uid
    except KeyError:
        # Let's check for 'node_types'
        pass

    if not data.get('node_types', None):
        return

    for ntype in data['node_types'].values():
        if ntype['derived_from'] != 'tosca.nodes.nfv.VNF':
            continue
        try:
            desc_id = ntype['properties']['descriptor_id']
            if desc_id.get('constraints', None):
                for constraint in desc_id.get('constraints'):
                    if constraint.get('valid_values', None):
                        constraint['valid_values'] = [uid]
            if desc_id.get('default', None):
                desc_id['default'] = uid
        except KeyError:
            # Let's check next node_type
            pass


def create_csar_with_unique_vnfd_id(csar_dir):
    """Create CSAR file from a directory structure

    For various tests it is necessary to have a CSAR having unique vnfd id.
    This function reads a directory structure, updates vnfd id in yaml files
    and creates a temporary CSAR zip file.

    :returns:
        - csar_file_name
        - vnfd_id
    """
    current_dir = os.path.dirname(os.path.abspath(__file__))
    csar_dir = os.path.join(current_dir, "../../", csar_dir)
    unique_id = uuidutils.generate_uuid()
    tempfd, tempname = tempfile.mkstemp(suffix=".zip",
        dir=os.path.dirname(csar_dir))
    os.close(tempfd)
    common_dir = os.path.join(csar_dir, "../common/")
    zcsar = zipfile.ZipFile(tempname, 'w')

    artifact_files = []
    for (dpath, _, fnames) in os.walk(csar_dir):
        if not fnames:
            continue
        for fname in fnames:
            if fname == 'TOSCA.meta' or fname.endswith('.mf'):
                src_file = os.path.join(dpath, fname)
                with open(src_file, 'rb') as f:
                    artifacts_data = f.read()
                artifacts_data_split = re.split(b'\n\n+', artifacts_data)
                for data in artifacts_data_split:
                    if re.findall(b'.?Algorithm:.?|.?Hash:.?', data):
                        artifact_data_dict = yaml.safe_load(data)
                        artifact_files.append(
                            artifact_data_dict['Source']
                            if 'Source' in artifact_data_dict.keys()
                            else artifact_data_dict['Name'])
    artifact_files = list(set(artifact_files))

    for (dpath, _, fnames) in os.walk(csar_dir):
        if not fnames:
            continue
        for fname in fnames:
            src_file = os.path.join(dpath, fname)
            dst_file = os.path.relpath(os.path.join(dpath, fname), csar_dir)
            if fname.endswith('.yaml') or fname.endswith('.yml'):
                if dst_file not in artifact_files:
                    with open(src_file, 'rb') as yfile:
                        data = yaml.safe_load(yfile)
                        _update_unique_id_in_yaml(data, unique_id)
                        zcsar.writestr(dst_file, yaml.dump(
                            data, default_flow_style=False,
                            allow_unicode=True))
                else:
                    zcsar.write(src_file, dst_file)
            else:
                zcsar.write(src_file, dst_file)

    for (dpath, _, fnames) in os.walk(common_dir):
        if not fnames:
            continue
        if ('test_cnf' in csar_dir and
                re.search('images|kubernetes|Scripts', dpath)):
            continue
        for fname in fnames:
            src_file = os.path.join(dpath, fname)
            dst_file = os.path.relpath(os.path.join(dpath, fname), common_dir)
            zcsar.write(src_file, dst_file)

    zcsar.close()
    return tempname, unique_id


def create_csar_with_unique_artifact(csar_dir):
    unique_id = uuidutils.generate_uuid()
    tempfd, tempname = tempfile.mkstemp(suffix=".zip",
                                        dir=os.path.dirname(csar_dir))
    os.close(tempfd)
    common_artifact_dir = os.path.join(csar_dir, '../common_artifact')
    common_dir = os.path.join(csar_dir, '../common')
    zcsar = zipfile.ZipFile(tempname, 'w')

    artifact_files = []
    for (dpath, _, fnames) in os.walk(csar_dir):
        if not fnames:
            continue
        for fname in fnames:
            if fname == 'TOSCA.meta' or fname.endswith('.mf'):
                src_file = os.path.join(dpath, fname)
                with open(src_file, 'rb') as f:
                    artifacts_data = f.read()
                artifacts_data_split = re.split(b'\n\n+', artifacts_data)
                for data in artifacts_data_split:
                    if re.findall(b".?Algorithm:.?", data) and\
                            re.findall(b".?Hash:.?", data):
                        artifact_data_dict = yaml.safe_load(data)
                        artifact_files.append(
                            artifact_data_dict['Source']
                            if 'Source' in artifact_data_dict.keys()
                            else artifact_data_dict['Name'])
    artifact_files = list(set(artifact_files))

    for (dpath, _, fnames) in os.walk(common_artifact_dir):
        if not fnames:
            continue
        for fname in fnames:
            src_file = os.path.join(dpath, fname)
            dst_file = os.path.relpath(
                os.path.join(dpath, fname), common_artifact_dir)
            if fname.endswith('.yaml') or fname.endswith('.yml'):
                if dst_file not in artifact_files:
                    with open(src_file, 'rb') as yfile:
                        data = yaml.safe_load(yfile)
                        _update_unique_id_in_yaml(data, unique_id)
                        zcsar.writestr(dst_file, yaml.dump(
                            data, default_flow_style=False,
                            allow_unicode=True))
                else:
                    zcsar.write(src_file, dst_file)
            else:
                zcsar.write(src_file, dst_file)

    for (dpath, _, fnames) in os.walk(csar_dir):
        if not fnames:
            continue
        for fname in fnames:
            src_file = os.path.join(dpath, fname)
            dst_file = os.path.relpath(os.path.join(dpath, fname), csar_dir)
            zcsar.write(src_file, dst_file)

    for (dpath, _, fnames) in os.walk(common_dir):
        if not fnames:
            continue
        for fname in fnames:
            src_file = os.path.join(dpath, fname)
            dst_file = os.path.relpath(os.path.join(dpath, fname), common_dir)
            zcsar.write(src_file, dst_file)

    zcsar.close()

    return tempname


def copy_csar_files(fake_csar_path, csar_dir_name,
                    csar_without_tosca_meta=False, read_vnfd_only=False):
    """Copy csar directory to temporary directory

    :param fake_csar_path: a temporary directory in which csar data will be
                           copied.
    :param csar_dir_name: the relative path of csar directory with respect to
                          './tacker/tests/etc/samples/etsi/nfv' directory.
                          for ex. 'vnfpkgm1'.
    :param csar_without_tosca_meta: when set to 'True', it will only copy root
                                    level yaml file and image file.
    :param read_vnfd_only: when set to 'True', it won't copy the image file
                           from source directory.
    """
    current_dir = os.path.dirname(os.path.abspath(__file__))
    sample_vnf_package = os.path.join(current_dir, "etc/samples/etsi/nfv",
                                      csar_dir_name)
    shutil.copytree(sample_vnf_package, fake_csar_path)
    common_files_path = os.path.join(current_dir,
                                     "etc/samples/etsi/nfv/common/")

    if not read_vnfd_only:
        # Copying image file.
        shutil.copytree(os.path.join(common_files_path, "Files/"),
                        os.path.join(fake_csar_path, "Files/"))

    if csar_without_tosca_meta:
        return

    # Copying common vnfd files.
    tosca_definition_file_path = os.path.join(common_files_path,
                                              "Definitions/")
    for (dpath, _, fnames) in os.walk(tosca_definition_file_path):
        if not fnames:
            continue
        for fname in fnames:
            src_file = os.path.join(dpath, fname)
            shutil.copy(src_file, os.path.join(fake_csar_path,
                                               "Definitions"))


def copy_artifact_files(fake_csar_path, csar_dir_name,
                    csar_without_tosca_meta=False, read_vnfd_only=False):
    current_dir = os.path.dirname(os.path.abspath(__file__))
    sample_vnf_package = os.path.join(current_dir,
                                      "etc/samples/etsi/nfv",
                                      csar_dir_name)
    shutil.copytree(sample_vnf_package, fake_csar_path)
    common_files_path = os.path.join(current_dir,
                                     "etc/samples/etsi/nfv/")

    if not read_vnfd_only:
        # Copying image file.
        shutil.copytree(os.path.join(common_files_path, "Files/"),
                        os.path.join(fake_csar_path, "Files/"))
        shutil.copytree(os.path.join(common_files_path, "Scripts/"),
                        os.path.join(fake_csar_path, "Scripts/"))

    if csar_without_tosca_meta:
        return

    # Copying common vnfd files.
    tosca_definition_file_paths = [
        os.path.join(common_files_path, "common_artifact/Definitions/"),
        os.path.join(common_files_path, "common/Definitions/")
    ]
    for tosca_definition_file_path in tosca_definition_file_paths:
        for (dpath, _, fnames) in os.walk(tosca_definition_file_path):
            if not fnames:
                continue
            for fname in fnames:
                src_file = os.path.join(dpath, fname)
                if not os.path.exists(os.path.join(
                        fake_csar_path, "Definitions")):
                    os.mkdir(os.path.join(fake_csar_path, "Definitions"))
                os.mknod(os.path.join(fake_csar_path, "Definitions", fname))
                shutil.copyfile(src_file, os.path.join(
                    fake_csar_path, "Definitions", fname))


class AuthHandler(http.server.SimpleHTTPRequestHandler):
    '''Main class to present webpages and authentication.'''

    def do_AUTHHEAD(self):
        self.send_response(401)
        self.send_header('WWW-Authenticate', 'Basic realm=\"Test\"')
        self.send_header('Content-type', 'text/plain')
        self.end_headers()

    def do_GET(self):
        '''Present frontpage with user authentication.'''
        global key
        if 'Authorization' not in self.headers:
            http.server.SimpleHTTPRequestHandler.do_GET(self)
        elif self.headers.get('Authorization') is None:
            self.do_AUTHHEAD()
            self.wfile.write(bytes('no auth header received'))
        elif self.headers.get('Authorization') == 'Basic ' + base64.b64encode(
                b"username:password").decode("utf-8"):
            http.server.SimpleHTTPRequestHandler.do_GET(self)
        else:
            self.do_AUTHHEAD()
            self.wfile.write(bytes(self.headers.get('Authorization')))
            self.wfile.write(bytes('not authenticated'))


class StaticHttpFileHandler(object):

    def __init__(self, static_files_path):
        if os.path.isabs(static_files_path):
            web_dir = static_files_path
        else:
            web_dir = os.path.join(os.path.dirname(__file__),
                static_files_path)
        os.chdir(web_dir)
        server_address = ('127.0.0.1', 0)
        self.httpd = socketserver.TCPServer(server_address, AuthHandler)
        self.port = self.httpd.socket.getsockname()[1]

        thread = threading.Thread(target=self.httpd.serve_forever)
        thread.daemon = True
        thread.start()

    def stop(self):
        self.httpd.shutdown()
        self.httpd.server_close()



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.tests.functional.sol_kubernetes.vnflcm import base as vnflcm_base


class VnfLcmKubernetesHealTest(vnflcm_base.BaseVnfLcmKubernetesTest):

    @classmethod
    def setUpClass(cls):
        super(VnfLcmKubernetesHealTest, cls).setUpClass()
        vnf_package_id, cls.vnfd_id = \
            cls._create_and_upload_vnf_package(
                cls, cls.tacker_client, "test_cnf_heal",
                {"key": "sample_heal_functional"})
        cls.vnf_package_ids.append(vnf_package_id)

    @classmethod
    def tearDownClass(cls):
        super(VnfLcmKubernetesHealTest, cls).tearDownClass()

    def test_heal_cnf_with_sol002(self):
        """Test heal as per SOL002 for CNF

        This test will instantiate cnf. Heal API will be invoked as per SOL002
        i.e. with vnfcInstanceId, so that the specified vnfc instance is healed
        which includes Kubernetes resources (Pod and Deployment).
        """
        vnf_instance_name = "cnf_heal_with_sol002"
        vnf_instance_description = "cnf heal with sol002"
        # use def-files of singleton Pod and Deployment (replicas=2)
        inst_additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/deployment_heal_complex.yaml",
                "Files/kubernetes/pod_heal.yaml"]}
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "complex", vnf_instance_name,
            vnf_instance_description, inst_additional_param)
        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)

        # get vnfc_instance_id of heal target
        deployment_target_vnfc = None
        for vnfc_rsc in before_vnfc_rscs:
            compute_resource = vnfc_rsc['computeResource']
            rsc_kind = compute_resource['vimLevelResourceType']
            if rsc_kind == 'Pod':
                # target 1: Singleton Pod
                pod_target_vnfc = vnfc_rsc
            elif not deployment_target_vnfc:
                # target 2: Deployment's Pod
                deployment_target_vnfc = vnfc_rsc
            else:
                # not target: Deployment's remianing one
                deployment_not_target_vnfc = vnfc_rsc

        # test heal SOL-002 (partial heal)
        vnfc_instance_id = \
            [pod_target_vnfc['id'], deployment_target_vnfc['id']]
        after_vnfc_rscs = self._test_heal(vnf_instance, vnfc_instance_id)
        for vnfc_rsc in after_vnfc_rscs:
            after_pod_name = vnfc_rsc['computeResource']['resourceId']
            if vnfc_rsc['id'] == pod_target_vnfc['id']:
                # check stored pod name is not changed (Pod)
                after_resource = pod_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertEqual(after_pod_name, before_pod_name)
            elif vnfc_rsc['id'] == deployment_target_vnfc['id']:
                # check stored pod name is changed (Deployment)
                after_resource = deployment_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertNotEqual(after_pod_name, before_pod_name)
            else:
                # check stored pod name is not changed (not target)
                after_resource = deployment_not_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertEqual(after_pod_name, before_pod_name)
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])

    def test_heal_cnf_with_sol003(self):
        """Test heal as per SOL003 for CNF

        This test will instantiate cnf. Heal API will be invoked as per SOL003
        i.e. without passing vnfcInstanceId, so that the entire vnf is healed
        which includes Kubernetes resource (Deployment).
        """
        vnf_instance_name = "cnf_heal_with_sol003"
        vnf_instance_description = "cnf heal with sol003"
        # use def-files of Deployment (replicas=2)
        inst_additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/deployment_heal_simple.yaml"]}
        vnf_instance = self._create_and_instantiate_vnf_instance(
            self.vnfd_id, "simple", vnf_instance_name,
            vnf_instance_description, inst_additional_param)
        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)

        # test heal SOL-003 (entire heal)
        vnfc_instance_id = []
        after_vnfc_rscs = self._test_heal(vnf_instance, vnfc_instance_id)
        self.assertEqual(len(before_vnfc_rscs), len(after_vnfc_rscs))
        # check id and pod name (as computeResource.resourceId) is changed
        for before_vnfc_rsc in before_vnfc_rscs:
            for after_vnfc_rsc in after_vnfc_rscs:
                self.assertNotEqual(
                    before_vnfc_rsc['id'], after_vnfc_rsc['id'])
                self.assertNotEqual(
                    before_vnfc_rsc['computeResource']['resourceId'],
                    after_vnfc_rsc['computeResource']['resourceId'])
        # terminate vnf instance
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])



# Copyright (c) 2012 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import routes as routes_mapper
from urllib import parse as urlparse
import webob
import webob.dec
import webob.exc

from tacker.api import extensions
from tacker.api.v1 import attributes
from tacker import wsgi


class Index(wsgi.Application):
    def __init__(self, resources):
        self.resources = resources

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        metadata = {}

        layout = []
        for name, collection in (self.resources).items():
            href = urlparse.urljoin(req.path_url, collection)
            resource = {'name': name,
                        'collection': collection,
                        'links': [{'rel': 'self',
                                   'href': href}]}
            layout.append(resource)
        response = dict(resources=layout)
        content_type = req.best_match_content_type()
        body = wsgi.Serializer(metadata=metadata).serialize(response,
                                                            content_type)
        return webob.Response(body=body, content_type=content_type)


class APIRouter(wsgi.Router):

    @classmethod
    def factory(cls, global_config, **local_config):
        return cls(**local_config)

    def __init__(self, **local_config):
        mapper = routes_mapper.Mapper()
        ext_mgr = extensions.ExtensionManager.get_instance()
        ext_mgr.extend_resources("1.0", attributes.RESOURCE_ATTRIBUTE_MAP)
        super(APIRouter, self).__init__(mapper)



# Copyright 2016 Brocade Communications System, Inc.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from toscaparser import tosca_template
import yaml

from tacker.common import utils
from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tosca import utils as toscautils

CONF = cfg.CONF


class VnfTestToscaMultipleVDU(base.BaseTackerTest):
    def test_create_delete_tosca_vnf_with_multiple_vdus(self):
        vnf_name = 'test_tosca_vnf_with_multiple_vdus'
        vnfd_file = 'sample-tosca-vnfd-multi-vdu.yaml'
        vnfd_instance, vnf_instance, tosca_dict = self.vnfd_and_vnf_create(
            vnfd_file, vnf_name)

        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(vnf_id,
                                   constants.VNF_CIRROS_CREATE_TIMEOUT,
                                   constants.ACTIVE_SLEEP_TIME)
        self.assertEqual('ACTIVE',
                         self.client.show_vnf(vnf_id)['vnf']['status'])
        self.validate_vnf_instance(vnfd_instance, vnf_instance)

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE,
            evt_constants.PENDING_CREATE, cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.ACTIVE)

        # Validate mgmt_ip_address with input yaml file
        mgmt_ip_address = self.client.show_vnf(vnf_id)['vnf'][
            'mgmt_ip_address']
        self.assertIsNotNone(mgmt_ip_address)
        mgmt_dict = yaml.safe_load(str(mgmt_ip_address))

        toscautils.updateimports(tosca_dict)

        tosca = tosca_template.ToscaTemplate(parsed_params={}, a_file=False,
                          yaml_dict_tpl=tosca_dict)

        vdus = toscautils.findvdus(tosca)

        self.assertEqual(len(vdus), len(mgmt_dict.keys()))
        for vdu in vdus:
            self.assertIsNotNone(mgmt_dict[vdu.name])
            self.assertEqual(True, utils.is_valid_ipv4(mgmt_dict[vdu.name]))

        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete of test_vnf_with_multiple_vdus failed"

        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)
